# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Pipeline which computes the number of examples in a dataset."""

import collections
import dataclasses
import functools
import itertools
import os
import pprint
from typing import cast, Dict, List, Optional, Type, Union

from etils import epath
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core import lazy_imports_lib
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import splits as split_lib
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.proto import dataset_info_pb2
from tensorflow_datasets.core.utils.lazy_imports_utils import array_record_module

from google.protobuf import json_format

_SplitFilesDict = Dict[str, List[naming.FilenameInfo]]


@dataclasses.dataclass
class _ShardInfo:
  """Metadata computed for each shard."""

  file_info: naming.FilenameInfo
  num_examples: int
  bytes_size: int


def _enrich_filename_template(
    filename_template: naming.ShardedFileTemplate,
    files_per_split: _SplitFilesDict,
) -> naming.ShardedFileTemplate:
  """Overrides the template's dataset name and suffix based on the filenames."""
  found_dataset_names = set()
  found_filetype_suffixes = set()
  for file_infos in files_per_split.values():
    for file_info in file_infos:
      found_dataset_names.add(file_info.dataset_name)
      found_filetype_suffixes.add(file_info.filetype_suffix)

  dataset_name = None
  if len(found_dataset_names) == 1:
    dataset_name = next(iter(found_dataset_names))
  elif len(found_dataset_names) > 1:
    raise ValueError(f'Found multiple dataset names: {found_dataset_names}')

  filetype_suffix = None
  if len(found_filetype_suffixes) == 1:
    filetype_suffix = next(iter(found_filetype_suffixes))
  if len(found_filetype_suffixes) > 1:
    raise ValueError(
        f'Found multiple filetype suffixes: {found_filetype_suffixes}'
    )

  return filename_template.replace(
      dataset_name=dataset_name, filetype_suffix=filetype_suffix
  )


def compute_split_info_from_directory(
    *,
    out_dir: Optional[epath.PathLike] = None,
    data_dir: epath.PathLike,
    filename_template: Union[None, str, naming.ShardedFileTemplate] = None,
) -> List[split_lib.SplitInfo]:
  """Compute the split info for the splits in the given data dir.

  Arguments:
    out_dir: directory where to save the metadata. It should be available from
      the apache beam workers. If not set, apache beam won't be used (only
      available with some file formats).
    data_dir: directory where the data is.
    filename_template: the template to which the data files correspond. If None,
      then the default template is used.

  Returns:
    list of split infos for the splits in the given data dir.
  """
  data_dir = epath.Path(data_dir)
  if filename_template is None:
    filename_template = naming.ShardedFileTemplate(data_dir=data_dir)
  elif isinstance(filename_template, str):
    filename_template = naming.ShardedFileTemplate(
        template=filename_template, data_dir=data_dir
    )
  filename_template = filename_template.replace(data_dir=data_dir)

  if (
      filename_template.dataset_name is None
      or filename_template.filetype_suffix is None
  ):
    # Get the dataset name and filetype suffix from the files in the data dir.
    files_per_split = _extract_split_files(filename_template)
    filename_template = _enrich_filename_template(
        filename_template=filename_template, files_per_split=files_per_split
    )
  return compute_split_info(
      out_dir=out_dir, filename_template=filename_template
  )


def compute_split_info(
    *,
    out_dir: Optional[epath.PathLike] = None,
    filename_template: naming.ShardedFileTemplate,
) -> List[split_lib.SplitInfo]:
  """Compute the split info on the given files.

  Compute the split info (num shards, num examples,...) metadata required
  by `tfds.folder_dataset.write_metadata`.

  See documentation for usage:
  https://www.tensorflow.org/datasets/external_tfrecord

  Args:
    out_dir: Output directory where to save the metadata. It should be available
      from the apache beam workers. If not set, apache beam won't be used (only
      available with some file formats).
    filename_template: filename template of the splits. The template should have
      set the data_dir because this is used to compute the split info.

  Returns:
    split_infos: The list of `tfds.core.SplitInfo`.
  """
  # Auto-detect the splits from the files
  split_files = _extract_split_files(filename_template)
  print('Auto-detected splits:')
  for split_name, file_infos in split_files.items():
    print(f' * {split_name}: {file_infos[0].num_shards} shards')

  # Launch the beam pipeline to compute the split info
  if out_dir is not None:
    split_infos = _compute_split_statistics_beam(
        split_files=split_files,
        out_dir=out_dir,
        filename_template=filename_template,
    )
  else:
    split_infos = _compute_split_statistics(
        split_files=split_files,
        filename_template=filename_template,
    )

  print('Computed split infos: ')
  pprint.pprint(split_infos)

  return split_infos


def _extract_split_files(
    filename_template: naming.ShardedFileTemplate,
) -> _SplitFilesDict:
  """Extract the files."""
  files = sorted(filename_template.data_dir.iterdir())
  file_infos = [filename_template.parse_filename_info(f.name) for f in files]
  file_infos = [f for f in file_infos if f is not None]
  if not file_infos:
    raise ValueError(
        f'No example files detected in {filename_template.data_dir}. '
        f'Make sure to follow the pattern: {filename_template.template}'
    )

  files_without_splits = []
  split_files = collections.defaultdict(list)
  for file_info in file_infos:
    if file_info.split is not None:
      split_files[file_info.split].append(file_info)
    else:
      files_without_splits.append(file_info)
  if files_without_splits:
    raise ValueError(
        f'Some matched files did not specify the split: {files_without_splits}'
    )

  return split_files


def _assert_split_is_consistent(file_infos: List[naming.FilenameInfo]) -> None:
  # Use unpack syntax on set to implicitly check that all values are the same
  (_,) = {f.split for f in file_infos}

  # Check that all the file-info from the given split are consistent
  # (no missing file)
  shard_ids = sorted(f.shard_index for f in file_infos)
  (num_shards,) = {f.num_shards for f in file_infos}
  if num_shards:
    assert shard_ids == list(range(num_shards)), 'Missing shard files.'


def _compute_split_statistics(
    *,
    split_files: _SplitFilesDict,
    filename_template: naming.ShardedFileTemplate,
) -> List[split_lib.SplitInfo]:
  """Computes and returns the split statistics."""

  adapter = None
  for _, file_infos in split_files.items():
    _assert_split_is_consistent(file_infos)
    if adapter is None:
      (file_suffix,) = {f.filetype_suffix for f in file_infos}
      file_format = file_adapters.file_format_from_suffix(file_suffix)
      adapter = file_adapters.ADAPTER_FOR_FORMAT[file_format]

  # Compute all shard info in parallel
  split_to_shard_infos = cast(
      Dict[str, List[_ShardInfo]],
      utils.tree.parallel_map(
          functools.partial(
              _process_shard,
              data_dir=filename_template.data_dir,
              adapter=adapter,
          ),
          split_files,
          report_progress=True,
      ),
  )
  # Create the SplitInfo for all splits
  return [
      _merge_shard_info(shard_infos=si, filename_template=filename_template)
      for si in split_to_shard_infos.values()
  ]


def _compute_split_statistics_beam(
    *,
    split_files: _SplitFilesDict,
    out_dir: epath.PathLike,
    filename_template: naming.ShardedFileTemplate,
) -> List[split_lib.SplitInfo]:
  """Compute statistics."""
  out_dir = epath.Path(out_dir)

  assert out_dir.exists(), f'{out_dir} does not exist'

  beam = lazy_imports_lib.lazy_imports.apache_beam

  # Launch the beam pipeline computation
  runner = None
  # Create the global pipeline object common for all splits
  # Disable type_hint as it doesn't works with typing.Protocol
  beam_options = beam.options.pipeline_options.PipelineOptions()
  beam_options.view_as(
      beam.options.pipeline_options.TypeOptions
  ).pipeline_type_check = False
  with beam.Pipeline(runner=runner, options=beam_options) as pipeline:
    for split_name, file_infos in split_files.items():
      _ = (
          pipeline
          | split_name
          >> _process_split(  # pylint: disable=no-value-for-parameter
              filename_template=filename_template,
              out_dir=out_dir,
              file_infos=file_infos,  # pytype: disable=missing-parameter
          )
      )

  # After the files have been computed
  return [
      _split_info_from_path(
          filename_template.replace(data_dir=out_dir, split=split)
      )
      for split in split_files
  ]


@lazy_imports_lib.beam_ptransform_fn
def _process_split(
    pipeline,
    *,
    filename_template: naming.ShardedFileTemplate,
    out_dir: epath.Path,
    file_infos: List[naming.FilenameInfo],
):
  """Process a single split."""
  beam = lazy_imports_lib.lazy_imports.apache_beam

  # Use unpack syntax on set to implicitly check that all values are the same
  (split_name,) = {f.split for f in file_infos}

  # Check that all the file-info from the given split are consistent
  # (no missing file)
  shard_ids = sorted(f.shard_index for f in file_infos)
  (num_shards,) = {f.num_shards for f in file_infos}
  assert shard_ids == list(range(num_shards)), 'Missing shard files.'

  # Check that the file extension is correct.
  (file_suffix,) = {f.filetype_suffix for f in file_infos}
  file_format = file_adapters.file_format_from_suffix(file_suffix)
  adapter = file_adapters.ADAPTER_FOR_FORMAT[file_format]
  data_dir = epath.Path(filename_template.data_dir)

  # Build the pipeline to process one split
  return (
      pipeline
      | beam.Create(file_infos)
      | beam.Map(_process_shard, data_dir=data_dir, adapter=adapter)
      # Group everything in a single elem (_ShardInfo -> List[_ShardInfo])
      | _group_all()  # pytype: disable=missing-parameter  # pylint: disable=no-value-for-parameter
      | beam.Map(_merge_shard_info, filename_template=filename_template)
      | beam.Map(_split_info_to_json_str)
      | beam.io.WriteToText(  # pytype: disable=missing-parameter
          os.fspath(out_dir / _out_filename(split_name)),
          num_shards=1,
          shard_name_template='',
      )
  )


@lazy_imports_lib.beam_ptransform_fn
def _group_all(pipeline):
  beam = lazy_imports_lib.lazy_imports.apache_beam
  # We do not use CombineGlobally(_process_shard) as it might be called
  # recursively. We want `_process_shard` to be called only once on the full
  # collection.
  return (
      pipeline | beam.GroupBy(lambda x: None) | beam.MapTuple(lambda key, x: x)
  )


def _process_shard(
    file_info: naming.FilenameInfo,
    *,
    data_dir: epath.Path,
    adapter: Type[file_adapters.FileAdapter],
) -> _ShardInfo:
  """Process a single `.tfrecord` file."""
  # Load the ds
  ds = adapter.make_tf_data(data_dir / str(file_info))
  num_examples = 0
  bytes_size = 0
  for ex in ds:
    num_examples += 1
    bytes_size += len(ex._numpy())  # pylint: disable=protected-access
  return _ShardInfo(
      file_info=file_info,
      num_examples=num_examples,
      bytes_size=bytes_size,
  )


def _merge_shard_info(
    shard_infos: List[_ShardInfo],
    filename_template: naming.ShardedFileTemplate,
) -> split_lib.SplitInfo:
  """Merge all shard info from one splits and returns the SplitInfo.

  Args:
    shard_infos: The shard infos from all shards.
    filename_template: filename template of the splits.

  Returns:
    The json SplitInfo proto
  """
  (split_name,) = {s.file_info.split for s in shard_infos}
  shard_infos = sorted(shard_infos, key=lambda s: s.file_info.shard_index)
  filename_template = filename_template.replace(split=split_name)
  return split_lib.SplitInfo(
      name=split_name,
      shard_lengths=[s.num_examples for s in shard_infos],
      num_bytes=sum(s.bytes_size for s in shard_infos),
      filename_template=filename_template,
  )


def _split_info_to_json_str(split_info: split_lib.SplitInfo) -> str:
  proto = split_info.to_proto()
  return json_format.MessageToJson(proto, sort_keys=True)


def _out_filename(split_name: str) -> str:
  return f'{split_name}-info.json'


def _split_info_from_path(
    filename_template: naming.ShardedFileTemplate,
) -> split_lib.SplitInfo:
  """Load the split info from the path."""
  path = filename_template.data_dir / _out_filename(filename_template.split)
  json_str = path.read_text()
  proto = json_format.Parse(json_str, dataset_info_pb2.SplitInfo())
  return split_lib.SplitInfo.from_proto(
      proto, filename_template=filename_template.replace(split=proto.name)
  )


def split_infos_from_path(
    split_names: List[str],
    filename_template: naming.ShardedFileTemplate,
) -> List[split_lib.SplitInfo]:
  """Restore the split info from a directory."""
  return [
      _split_info_from_path(
          filename_template=filename_template.replace(split=split_name)
      )
      for split_name in split_names
  ]



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""GZTAN dataset."""

import os

from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

_CITATION = """
@misc{tzanetakis_essl_cook_2001,
author    = "Tzanetakis, George and Essl, Georg and Cook, Perry",
title     = "Automatic Musical Genre Classification Of Audio Signals",
url       = "http://ismir2001.ismir.net/pdf/tzanetakis.pdf",
publisher = "The International Society for Music Information Retrieval",
year      = "2001"
}
"""

_DESCRIPTION = """
The dataset consists of 1000 audio tracks each 30 seconds long.
It contains 10 genres, each represented by 100 tracks.
The tracks are all 22050Hz Mono 16-bit audio files in .wav format.

The genres are:

* blues
* classical
* country
* disco
* hiphop
* jazz
* metal
* pop
* reggae
* rock

"""

_DOWNLOAD_URL = "http://opihi.cs.uvic.ca/sound/genres.tar.gz"
_HOMEPAGE_URL = "http://marsyas.info/index.html"

_CLASS_LABELS = [
    "blues",
    "classical",
    "country",
    "disco",
    "hiphop",
    "jazz",
    "metal",
    "pop",
    "reggae",
    "rock",
]


class GTZAN(tfds.core.GeneratorBasedBuilder):
  """GTZAN Dataset."""

  VERSION = tfds.core.Version("1.0.0")

  def _info(self):
    return tfds.core.DatasetInfo(
        builder=self,
        description=_DESCRIPTION,
        features=tfds.features.FeaturesDict({
            "audio": tfds.features.Audio(file_format="wav", sample_rate=22050),
            "label": tfds.features.ClassLabel(names=_CLASS_LABELS),
            "audio/filename": tfds.features.Text(),
        }),
        supervised_keys=("audio", "label"),
        homepage=_HOMEPAGE_URL,
        citation=_CITATION,
    )

  def _split_generators(self, dl_manager):
    """Returns SplitGenerators."""
    dl_paths = dl_manager.download_and_extract({"genres": _DOWNLOAD_URL})
    path = os.path.join(dl_paths["genres"], "genres")
    # There is no predefined train/val/test split for this dataset.
    return [
        tfds.core.SplitGenerator(
            name=tfds.Split.TRAIN, gen_kwargs={"path": path}
        ),
    ]

  def _generate_examples(self, path):
    """Yields examples.

    Args:
       path: Path of the downloaded and extracted directory

    Yields:
       Next examples
    """
    for root, _, file_name in tf.io.gfile.walk(path):
      for fname in file_name:
        if fname.endswith(".wav"):  # select only .wav files
          # Each .wav file has name in the format of <genre>.<number>.wav
          label = fname.split(".")[0]
          key = fname
          example = {
              "audio": os.path.join(root, fname),
              "label": label,
              "audio/filename": fname,
          }
          yield key, example



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for generating the C4 dataset."""

import dataclasses
import functools
import gzip
import hashlib
import heapq
import io
import re
import threading
from typing import Collection, Iterable, Mapping, Optional, Sequence, Tuple

from absl import logging
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds

# WET file constants
_PAGE_DELIMITER = "WARC/1.0"
_URL_KEY = "WARC-Target-URI:"
_URL_DATE = "WARC-Date:"
_CONTENT_TYPE = "Content-Type:"
_CONTENT_LEN = "Content-Length:"
_METADATA_PREFIXES = ("WARC", "CONTENT-", "Content-")

# Filters
_MIN_WORDS_PER_LINE = 5
_MIN_NUM_SENTENCES = 3
_MAX_WORD_LENGTH = 1000
_END_MARKS = (".", "?", "!", '"')
_ELLIPSIS = "..."
_POLICY_SUBSTRINGS = [
    "terms of use",
    "privacy policy",
    "cookie policy",
    "uses cookies",
    "use of cookies",
    "use cookies",
]

# Memoized sentence tokenizer.
_SENTENCE_TOKENIZER = None

UNKNOWN_LANGUAGE = "und"


@dataclasses.dataclass
class PageFeatures:
  url: str = ""
  normalized_url: str = ""
  text: str = ""
  timestamp: str = ""
  content_length: str = ""
  content_type: str = ""
  language: Optional[str] = None


def get_counter_inc_fn(namespace):
  def counter_inc_fn(counter, amt=1):
    tfds.core.lazy_imports.apache_beam.metrics.Metrics.counter(
        namespace, counter
    ).inc(amt)

  return counter_inc_fn


def get_hashed_url_filter_fn(predicate_fn):
  def filter_fn(page):
    url = page.normalized_url
    val = int(
        hashlib.md5(tf.compat.as_text(url).encode("utf-8")).hexdigest(), 16
    )
    return predicate_fn(val)

  return filter_fn


_nltk_lock = threading.Lock()


def _load_sentence_tokenizer():
  """Returns a sentence tokenization function."""
  nltk = tfds.core.lazy_imports.nltk
  # Lock to avoid a race-condition in the creation of the download directory.
  with _nltk_lock:
    nltk.download("punkt")
    return nltk.data.load("nltk:tokenizers/punkt/english.pickle")




def _get_sentences(text):
  global _SENTENCE_TOKENIZER
  if not _SENTENCE_TOKENIZER:
    _SENTENCE_TOKENIZER = _load_sentence_tokenizer()
  return list(_SENTENCE_TOKENIZER.tokenize(tf.compat.as_text(text)))


# Global lock used for language detection modules that aren't threadsafe.
langdetect_lock = threading.Lock()


def detect_english(page: PageFeatures, min_probability=0.99):
  """Yields page iff text is 'en' with at least `min_probability`."""
  text = page.text

  counter_inc_fn = get_counter_inc_fn("english-filter")

  langdetect = tfds.core.lazy_imports.langdetect
  # Make langdetect predictions deterministic.
  with langdetect_lock:
    langdetect.DetectorFactory.seed = 0
    try:
      predictions = langdetect.detect_langs(text)
    except langdetect.lang_detect_exception.LangDetectException:
      counter_inc_fn("filtered:langdetect_exception")
      return
  if not predictions:
    counter_inc_fn("filtered:no_predictions")
    return
  best_prediction = predictions[0]
  if best_prediction.prob < min_probability:
    counter_inc_fn("filtered:low_confidence")
    return
  if best_prediction.lang != "en":
    counter_inc_fn("filtered:ignored_language")
    counter_inc_fn("filtered:ignored_language-%s" % (best_prediction.lang,))
    return

  counter_inc_fn("passed")
  yield dataclasses.replace(page, language="en")


def detect_languages(pages, valid_languages):
  """Predicts page language using cld3 and adds to features."""

  beam = tfds.core.lazy_imports.apache_beam

  class _PredictLanguageFn(beam.DoFn):
    """Predicts page's language using cld3 and adds to features."""

    def __init__(self, valid_languages, min_probability=0.95):
      self._valid_languages = set(valid_languages)
      self._counter_inc_fn = get_counter_inc_fn("language-filter")
      self._min_probability = min_probability

    def start_bundle(self):
      with langdetect_lock:
        self._detector = tfds.core.lazy_imports.gcld3.NNetLanguageIdentifier(
            # CLD3 is not expected to work well on very short documents.
            min_num_bytes=100,
            max_num_bytes=10000,
        )

    def process(self, page: PageFeatures):
      with langdetect_lock:
        result = self._detector.FindLanguage(page.text)
      if not result.is_reliable:
        self._counter_inc_fn("filtered:no_predictions")
        lang = UNKNOWN_LANGUAGE
      elif result.probability < self._min_probability:
        self._counter_inc_fn("filtered:low_confidence")
        lang = UNKNOWN_LANGUAGE
      else:
        lang = result.language
        if lang not in self._valid_languages:
          self._counter_inc_fn("filtered:ignored_language")
          return
      self._counter_inc_fn("passed")
      self._counter_inc_fn("passed:%s" % lang)
      yield dataclasses.replace(page, language=lang)

  return pages | beam.ParDo(_PredictLanguageFn(valid_languages=valid_languages))


def get_clean_page_fn():
  """Returns `clean_page` with pre-compiled badword and citation regexes."""
  # Used to filter citation from Wikipedia pages (among others).
  citation_regex = re.compile(r"\[\d*\]|\[edit\]|\[citation needed\]")
  return functools.partial(clean_page, citation_regex=citation_regex)


def clean_page(
    page: PageFeatures,
    citation_regex,
    counter_inc_fn=None,
    min_words_per_line=_MIN_WORDS_PER_LINE,
    min_num_sentences=_MIN_NUM_SENTENCES,
    max_word_length=_MAX_WORD_LENGTH,
    line_delimiter="\n",
) -> Iterable[PageFeatures]:
  """Cleans a CommonCrawl page, yielding nothing if it should be skipped.

  Cleaning removes lines with no end marks or with too few words. After line
  filtering, pages are filtered out if they have too few sentences based on a
  simple count of end marks.

  Args:
    page: the features of the page.
    citation_regex: Regex to use for finding Wikipedia-like citations to filter.
    counter_inc_fn: function, a function taking the name of a counter to be
      incremented and the (optional) amount. Defaults to a beam Metric counter.
    min_words_per_line: int, the minimum number of words a line needs to not be
      removed.
    min_num_sentences: int, the minimum number of sentences a page needs to not
      be skipped.
    max_word_length: int, the maximum number of characters allowed in a word.
      Lines containing a word with too many characters are removed.
    line_delimiter: str, the delimiter used to separate and join lines.

  Yields:
    The url and cleaned text for the page.
  """
  text = page.text

  if not counter_inc_fn:
    counter_inc_fn = get_counter_inc_fn("clean-page")

  lines = text.splitlines()
  valid_lines = []
  num_sentences = 0

  def line_has_too_long_word(line):
    for word in line.split():
      if len(word) > max_word_length:
        return True
    return False

  for line in lines:
    line = line.strip()
    if line_has_too_long_word(line):
      counter_inc_fn("line-filtered:too_long_word")
      continue
    line = citation_regex.sub("", line)
    if not line.endswith(_END_MARKS) or line.endswith(_ELLIPSIS):
      counter_inc_fn("line-filtered:no_endmark")
      continue
    if len(line.split()) < min_words_per_line:
      counter_inc_fn("line-filtered:too_short")
      continue
    line_lower = line.lower()
    # Remove documents which contain lorem ipsum
    if "lorem ipsum" in line_lower:
      counter_inc_fn("filtered:loremipsum")
      return
    # Remove "javascript must be enabled" notices
    if "javascript" in line_lower:
      counter_inc_fn("line-filtered:javascript")
      continue
    # Remove docs which probably contain javascript code
    if "{" in line:
      counter_inc_fn("filtered:squigglybracket")
      return
    # Remove policy lines
    if any(p in line_lower for p in _POLICY_SUBSTRINGS):
      counter_inc_fn("line-filtered:policy")
      continue
    num_sentences += len(_get_sentences(line))
    valid_lines.append(line)
    counter_inc_fn("line-passed")

  if num_sentences < min_num_sentences:
    counter_inc_fn("filtered:too_few_sentences")
    return
  counter_inc_fn("passed")
  yield dataclasses.replace(page, text=line_delimiter.join(valid_lines).strip())


def _hash_text(text):
  return hashlib.md5(tf.compat.as_text(text).encode("utf-8")).hexdigest()


def _emit_url_to_lines(
    page: PageFeatures, line_delimiter="\n"
) -> Iterable[Tuple[str, str]]:
  """Emits url to all (lower-cased, hashed) lines."""
  text = page.text
  for line in text.split(line_delimiter):
    yield _hash_text(line.strip().lower()), page.url


def _remove_lines_from_text(
    el, counter_inc_fn, min_num_sentences, line_delimiter="\n"
) -> Iterable[PageFeatures]:
  """Removes all lines from the page that do not match the given set of hashes.

  Process the result of a join containing a single value for 'features' and zero
  or more values for 'lines'. Each value in 'lines' is a lower-cased, hashed
  line that has been selected to keep.

  Args:
    el: `(string, {'pages': PageFeatures, 'lines': [string]})`, element
      containing the result of a join on key with both the page text and
      lower-cased, hashed lines to remove.
    counter_inc_fn: function, a function taking the name of a counter to be
      incremented and the (optional) amount.
    min_num_sentences: int, the minimum number of sentences a page needs to not
      be skipped.
    line_delimiter: str, the delimiter used to separate and join lines.

  Yields:
    The page features with lines removed from text.
  """
  url, join_values = el
  pages = join_values["pages"]

  assert len(pages) == 1, "Invalid page count (%d) for %s" % (len(pages), url)
  page = pages[0]
  text = page.text
  lines_to_keep = set(join_values["lines"])
  new_lines = []
  hashed_lines = set()
  for line in text.split(line_delimiter):
    hashed_line = _hash_text(line.strip().lower())
    if hashed_line not in lines_to_keep:
      counter_inc_fn("line-filtered:global_duplicate")
    elif hashed_line in hashed_lines:
      counter_inc_fn("line-filtered:local_duplicate")
    else:
      counter_inc_fn("line-passed")
      new_lines.append(line)
      hashed_lines.add(hashed_line)
  new_text = line_delimiter.join(new_lines)
  if not new_text:
    counter_inc_fn("filtered:empty")
    return
  if min_num_sentences and len(_get_sentences(new_text)) < min_num_sentences:
    counter_inc_fn("filtered:too_few_sentences")
    return
  counter_inc_fn("passed")
  yield dataclasses.replace(page, text=new_text)


def remove_duplicate_text(
    pages, min_num_sentences=_MIN_NUM_SENTENCES, line_delimiter="\n"
):
  """Utility to remove duplicate lines across text documents."""
  beam = tfds.core.lazy_imports.apache_beam

  # Select a single URL for each line in the input pages.
  # Hash before comparison to avoid biasing by domain.
  # line, [url]
  line_to_selected_url = (
      pages
      | beam.FlatMap(_emit_url_to_lines, line_delimiter=line_delimiter)
      | beam.combiners.Top.PerKey(1, key=_hash_text, reverse=True)
  )
  # url, line
  lines_to_keep = line_to_selected_url | beam.Map(lambda x: (x[1][0], x[0]))

  # Output: url, text
  final_docs = (
      {"pages": pages | beam.Map(lambda p: (p.url, p)), "lines": lines_to_keep}
      | "group_features_and_lines_by_url" >> beam.CoGroupByKey()
      | beam.FlatMap(
          _remove_lines_from_text,
          counter_inc_fn=get_counter_inc_fn("dedupe-lines"),
          min_num_sentences=min_num_sentences,
          line_delimiter=line_delimiter,
      )
  )

  return final_docs


def split_wet_file(wet_file_path, counter_inc_fn=None, line_delimiter="\n"):
  """Split a WET file into separate pages."""
  logging.info("Splitting file: %s", wet_file_path)
  if not counter_inc_fn:
    counter_inc_fn = get_counter_inc_fn("split-wet-file")
  counter_inc_fn("wet-file")

  def _validate_features(page):
    """Return True if page features are valid."""
    if not page.url:
      counter_inc_fn("filtered:no_url")
    elif not page.text:
      counter_inc_fn("filtered:no_content")
    elif not page.timestamp:
      counter_inc_fn("filtered:no_timestamp")
    else:
      counter_inc_fn("passed")
      return True
    return False

  with tf.io.gfile.GFile(wet_file_path, "rb") as f, gzip.GzipFile(
      fileobj=f
  ) as g:
    page = PageFeatures()
    for i, line in enumerate(io.TextIOWrapper(g, encoding="utf-8")):  # pytype: disable=wrong-arg-types
      line = line.strip()
      if not line:
        continue
      if line == _PAGE_DELIMITER:
        if i > 0 and _validate_features(page):
          yield page
        page = PageFeatures()

      if line.startswith(_URL_KEY):
        page.url = line[len(_URL_KEY) :].strip()
        page.normalized_url = normalize_url(line[len(_URL_KEY) :].strip())

      if line.startswith(_URL_DATE):
        page.timestamp = line[len(_URL_DATE) :].strip()

      if line.startswith(_CONTENT_TYPE):
        page.content_type = line[len(_CONTENT_TYPE) :].strip()

      if line.startswith(_CONTENT_LEN):
        page.content_length = line[len(_CONTENT_LEN) :].strip()

      if line.startswith(_METADATA_PREFIXES):
        continue

      if page.text:
        page.text += line_delimiter
      page.text += line

    if _validate_features(page):
      yield page


def select_newest_page(pages):
  """Deterministically return as random page."""
  counter_inc_fn = get_counter_inc_fn("duplicate-url-filter")

  cnt = 0
  selected_page = None
  for p in pages:
    cnt += 1
    if not selected_page or p.timestamp > selected_page.timestamp:
      selected_page = p
  counter_inc_fn("filtered", cnt - 1)
  counter_inc_fn("passed")
  return selected_page


def is_valid_length(page: PageFeatures, max_length=1.9e5):
  """Returns False iff page's text is too long."""
  counter_inc_fn = get_counter_inc_fn("too-long-filter")
  if len(page.text) > max_length:
    counter_inc_fn("filtered")
    return False
  counter_inc_fn("passed")
  return True


def is_realnews_domain(
    page: PageFeatures, realnews_domains: Mapping[str, Collection[str]]
):
  """Returns False iff page's (sub)domain is not allowed."""
  counter_inc_fn = get_counter_inc_fn("realnews-domain-filter")
  url = page.normalized_url
  ext = tfds.core.lazy_imports.tldextract.extract(url)
  main_domain = ext.domain + "." + ext.suffix
  if main_domain not in realnews_domains:
    counter_inc_fn("filtered:bad_domain")
    return False
  allowed_subdomains = realnews_domains[main_domain]
  if (
      isinstance(allowed_subdomains, list)
      and ext.subdomain not in allowed_subdomains
  ):
    counter_inc_fn("filtered:bad_subdomain")
    return False
  counter_inc_fn("passed")
  return True


def filter_by_webtextlike(el):
  """Yields only pages with a matching WebText-like URL."""
  counter_inc_fn = get_counter_inc_fn("webtextlike-filter")
  _, join_values = el
  pages = join_values["pages"]
  webtextlike = join_values["webtextlike_urls"]
  if not webtextlike:
    counter_inc_fn("filtered")
    return
  if not pages:
    counter_inc_fn("missing-page")
    return
  assert len(pages) == 1
  counter_inc_fn("passed")
  yield pages[0]


def normalize_url(url):
  url = tf.compat.as_text(url)
  url = re.sub(r"https?:\/\/(www\.)?", "", url)
  url = re.sub(r"\?(utm_|ref|feed).*", "", url)
  url = url.rstrip("/")
  return url


def get_badwords_filter_fn(
    badwords: Mapping[str, Sequence[str]], filter_fraction: float = 1.0
):
  """Filters pages at given rate that contain language-specific bad word(s)."""
  badwords_regex = {}
  for lang, words in badwords.items():
    words = [re.escape(w) for w in words]
    badwords_regex[lang] = (
        # For Japanese, Thai, and Chinese, do not require word separations.
        re.compile("|".join(words))
        if lang in ("ja", "th", "zh")
        # For other languages, match only when flanked by non-word chars.
        else re.compile(r"(?:\W|^)({})(?:\W|$)".format("|".join(words)))
    )

  filter_ratio = float.as_integer_ratio(filter_fraction)
  keep_badword_page = get_hashed_url_filter_fn(
      lambda x: x % filter_ratio[1] >= filter_ratio[0]
  )

  def badwords_filter(page):
    lang = page.language.split("-")[0]  # remove suffix if present

    if lang in badwords_regex:
      text = page.text
      badwords_found = badwords_regex[lang].search(text.lower())
      if badwords_found is not None:
        if keep_badword_page(page):
          get_counter_inc_fn("badwords-filter")("soft-passed")
          get_counter_inc_fn("badwords-filter-%s" % lang)("soft-passed")
          return True
        get_counter_inc_fn("badwords-filter")("filtered")
        get_counter_inc_fn("badwords-filter-%s" % lang)("filtered")
        return False
      get_counter_inc_fn("badwords-filter-%s" % lang)("passed")

    get_counter_inc_fn("badwords-filter")("passed")
    return True

  return badwords_filter


def paragraph_filter(
    page, min_paragraphs=3, min_paragraph_len=200, line_delimiter="\n"
):
  """Returns False iff a page has too few or too short paragraphs."""
  lines = page.text.split(line_delimiter)
  # Filter out docs that don't have at least three "paragraphs"
  # (lines >= `min_paragraph_len` chars).
  if (
      len(lines) < min_paragraphs
      or min(heapq.nlargest(3, [len(l) for l in lines])) < min_paragraph_len
  ):
    get_counter_inc_fn("paragraph-filter")("filtered")
    return False
  get_counter_inc_fn("paragraph-filter")("passed")
  return True



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Add HuggingFace Datasets.

Script to convert code for dataset in HuggingFace to tensorflow datasets

To convert a dataset from huggingface, first clone the git repo of
huggingface to local device.

To clone, copy and execute the following command in terminal

git clone https://github.com/huggingface/nlp.git

This will create a copy of the repo in local device. All datasets in
huggingface nlp will be copied.

"""

import argparse
import pathlib
import re

from absl import app
from absl.flags import argparse_flags
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.scripts.cli import new

TO_CONVERT = [
    # (pattern, replacement)
    # Order is important here for some replacements
    (r"from\s__future__\simport\sabsolute_import.*", r""),
    (r"from\s__future__\simport\sdivision.*", r""),
    (r"from\s__future__\simport\sprint_function.*", r""),
    (r"import\slogging", r"from absl import logging\n"),
    (
        r"import\snlp",
        (
            r"from tensorflow_datasets.core.utils.lazy_imports_utils import"
            r" tensorflow as tf\nimport tensorflow_datasets.public_api as"
            r" tfds\n"
        ),
    ),
    (r"with\sopen", r"with tf.io.gfile.GFile"),
    (r"encoding=\"utf-8\"", r"'r'"),
    (
        r"return\snlp\.DatasetInfo\(",
        r"    return tfds.core.DatasetInfo(\n        builder=self,\n",
    ),
    (r"nlp\.ClassLabel", r"tfds.features.ClassLabel"),
    (r"nlp\.Value\(\"string\"\)", r"tfds.features.Text()"),
    (r"nlp\.Value\(\"([\w\d]+)\"\)", r"tf.\1"),
    (r"nlp\.features", "tfds.features"),
    (
        r"features\s*=\s*nlp\.Features\(",
        r"features=tfds.features.FeaturesDict(",
    ),
    (r"dict\(", r"tfds.features.FeaturesDict("),
    (r"nlp.SplitGenerator", r"tfds.core.SplitGenerator"),
    (r"self\.config\.data_dir", r"dl_manager.manual_dir"),
    (r"self\.config", r"self.builder_config"),
    (r"nlp\.Split", r"tfds.Split"),
    (r"nlp", r"tfds.core"),
]


def _parse_flags(_) -> argparse.Namespace:
  """Command line flags."""
  parser = argparse_flags.ArgumentParser(
      prog="convert_dataset",
      description="Tool to add hugging face datasets",
  )
  parser.add_argument(
      "--nlp_path", type=pathlib.Path, help="Path of the nlp directory"
  )
  parser.add_argument(
      "--dataset_name",
      type=str,
      help="Name of dataset to be converted. Ex. esnli",
  )
  parser.add_argument(
      "--dataset_type",
      type=str,
      help="Type of dataset. Ex. text, question_answering",
  )
  return parser.parse_args()


def main(args: argparse.Namespace):
  create_dataset_files(
      nlp_path=args.nlp_path,
      dataset_name=args.dataset_name,
      dataset_type=args.dataset_type,
  )


def create_dataset_files(
    nlp_path: pathlib.Path,
    dataset_name: str,
    dataset_type: str,
) -> None:
  """Create template files."""
  #  Path of the converted dataset directory
  tfds_root_path = tfds.core.utils.tfds_write_path()
  dataset_dir = tfds_root_path / dataset_type
  if not dataset_dir.is_dir():
    raise ValueError(f"Invalid Dataset Type {dataset_type}")

  #  Create dataset timeplate files from new.py
  new.create_dataset_files(dataset_name=dataset_name, dataset_dir=dataset_dir)  # pytype: disable=wrong-arg-types  # gen-stub-imports

  #  Path of the dataset file
  nlp_datasets_path = nlp_path.expanduser() / "datasets"
  nlp_file_path = nlp_datasets_path / dataset_name / f"{dataset_name}.py"

  #  Path of converted dataset file
  converted_file_path = dataset_dir / dataset_name / f"{dataset_name}.py"

  #  Convert nlp dataset --> tfds
  convert_dataset_file(nlp_file_path, converted_file_path)


def convert_dataset_file(nlp_file_path, converted_file_path):
  """Convert nlp dataset."""
  print("Starting conversion of nlp file")
  compiled_patterns = [
      (re.compile(pattern), replacement) for pattern, replacement in TO_CONVERT
  ]

  nlp_file_contents = nlp_file_path.read_text()
  for pattern, replacement in compiled_patterns:
    nlp_file_contents = pattern.sub(replacement, nlp_file_contents)

  converted_file_path.write_text(nlp_file_contents)
  print("File successfully converted")
  print(nlp_file_contents)


if __name__ == "__main__":
  app.run(main, flags_parser=_parse_flags)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""UnifiedQA dataset."""

import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.text.unifiedqa import unifiedqa


class UnifiedQAAI2ScienceElementaryTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for ai2_science_elementary."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['ai2_science_elementary']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'ai2_science_elementary/train.tsv',
      'validation': 'ai2_science_elementary/dev.tsv',
      'test': 'ai2_science_elementary/test.tsv',
  }


class UnifiedQAAI2ScienceMiddleTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for ai2_science_middleai2_science_middle."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['ai2_science_middle']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'ai2_science_middle/train.tsv',
      'validation': 'ai2_science_middle/dev.tsv',
      'test': 'ai2_science_middle/test.tsv',
  }


class UnifiedQAAmbigQATest(tfds.testing.DatasetBuilderTestCase):
  """Tests for ambigqa."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['ambigqa']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'ambigqa/train.tsv',
      'validation': 'ambigqa/dev.tsv',
  }


class UnifiedQAARCEasyTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for arc_easy."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['arc_easy']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'arc_easy/train.tsv',
      'validation': 'arc_easy/dev.tsv',
      'test': 'arc_easy/test.tsv',
  }


class UnifiedQAARCEasyDevTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for arc_easy_dev."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['arc_easy_dev']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'arc_easy_dev/train.tsv',
      'validation': 'arc_easy_dev/dev.tsv',
      'test': 'arc_easy_dev/test.tsv',
  }


class UnifiedQAARCEasyIRTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for arc_easy_with_ir."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['arc_easy_with_ir']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'arc_easy_with_ir/train.tsv',
      'validation': 'arc_easy_with_ir/dev.tsv',
      'test': 'arc_easy_with_ir/test.tsv',
  }


class UnifiedQAARCEasyIRDevTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for arc_easy_with_ir_dev."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['arc_easy_with_ir_dev']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'arc_easy_with_ir_dev/train.tsv',
      'validation': 'arc_easy_with_ir_dev/dev.tsv',
      'test': 'arc_easy_with_ir_dev/test.tsv',
  }


class UnifiedQAARCHardTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for arc_hard."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['arc_hard']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'arc_hard/train.tsv',
      'validation': 'arc_hard/dev.tsv',
      'test': 'arc_hard/test.tsv',
  }


class UnifiedQAARCHardDevTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for arc_hard_dev."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['arc_hard_dev']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'arc_hard_dev/train.tsv',
      'validation': 'arc_hard_dev/dev.tsv',
      'test': 'arc_hard_dev/test.tsv',
  }


class UnifiedQAARCHardIRTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for arc_hard_with_ir."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['arc_hard_with_ir']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'arc_hard_with_ir/train.tsv',
      'validation': 'arc_hard_with_ir/dev.tsv',
      'test': 'arc_hard_with_ir/test.tsv',
  }


class UnifiedQAARCHardIRDevTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for arc_hard_with_ir_dev."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['arc_hard_with_ir_dev']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'arc_hard_with_ir_dev/train.tsv',
      'validation': 'arc_hard_with_ir_dev/dev.tsv',
      'test': 'arc_hard_with_ir_dev/test.tsv',
  }


class UnifiedQABoolQTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for boolq."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['boolq']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'boolq/train.tsv',
      'validation': 'boolq/dev.tsv',
  }


class UnifiedQABoolQNPTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for boolq_np."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['boolq_np']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'boolq_np/train.tsv',
      'validation': 'boolq_np/dev.tsv',
  }


class UnifiedQACQATest(tfds.testing.DatasetBuilderTestCase):
  """Tests for commonsenseqa."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['commonsenseqa']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'commonsenseqa/train.tsv',
      'validation': 'commonsenseqa/dev.tsv',
      'test': 'commonsenseqa/test.tsv',
  }


class UnifiedQACQATestTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for commonsenseqa_test."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['commonsenseqa_test']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'commonsenseqa_test/train.tsv',
      'validation': 'commonsenseqa_test/dev.tsv',
      'test': 'commonsenseqa_test/test.tsv',
  }


class UnifiedQAContrastSetsBoolQTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for contrast_sets_boolq."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['contrast_sets_boolq']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'contrast_sets_boolq/train.tsv',
      'validation': 'contrast_sets_boolq/dev.tsv',
  }


class UnifiedQAContrastSetsDROPTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for contrast_sets_drop."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['contrast_sets_drop']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'contrast_sets_drop/train.tsv',
      'validation': 'contrast_sets_drop/dev.tsv',
  }


class UnifiedQAContrastSetsQuorefTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for contrast_sets_quoref."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['contrast_sets_quoref']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'contrast_sets_quoref/train.tsv',
      'validation': 'contrast_sets_quoref/dev.tsv',
  }


class UnifiedQAContrastSetsROPESTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for contrast_sets_ropes."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['contrast_sets_ropes']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'contrast_sets_ropes/train.tsv',
      'validation': 'contrast_sets_ropes/dev.tsv',
  }


class UnifiedQADROPTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for drop."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['drop']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'drop/train.tsv',
      'validation': 'drop/dev.tsv',
  }


class UnifiedQAMCTestTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for mctest."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['mctest']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'mctest/train.tsv',
      'validation': 'mctest/dev.tsv',
  }


class UnifiedQAMCTestCorrectedTheSeparatorTest(
    tfds.testing.DatasetBuilderTestCase
):
  """Tests for mctest_corrected_the_separator."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['mctest_corrected_the_separator']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'mctest_corrected_the_separator/train.tsv',
      'validation': 'mctest_corrected_the_separator/dev.tsv',
  }


class UnifiedQAMultiRCTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for multirc."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['multirc']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'multirc/train.tsv',
      'validation': 'multirc/dev.tsv',
  }


class UnifiedQANarrativeQATest(tfds.testing.DatasetBuilderTestCase):
  """Tests for narrativeqa."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['narrativeqa']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'narrativeqa/train.tsv',
      'validation': 'narrativeqa/dev.tsv',
      'test': 'narrativeqa/test.tsv',
  }


class UnifiedQANarrativeQADevTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for narrativeqa_dev."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['narrativeqa_dev']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'narrativeqa_dev/train.tsv',
      'validation': 'narrativeqa_dev/dev.tsv',
      'test': 'narrativeqa_dev/test.tsv',
  }


class UnifiedQANatQATest(tfds.testing.DatasetBuilderTestCase):
  """Tests for natural_questions."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['natural_questions']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'natural_questions/train.tsv',
      'validation': 'natural_questions/dev.tsv',
  }


class UnifiedQANatQADirectAnsTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for natural_questions_direct_ans."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['natural_questions_direct_ans']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'natural_questions_direct_ans/train.tsv',
      'validation': 'natural_questions_direct_ans/dev.tsv',
      'test': 'natural_questions_direct_ans/test.tsv',
  }


class UnifiedQANatQADirectAnsTestTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for natural_questions_direct_ans_test."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['natural_questions_direct_ans_test']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'natural_questions_direct_ans_test/train.tsv',
      'validation': 'natural_questions_direct_ans_test/dev.tsv',
      'test': 'natural_questions_direct_ans_test/test.tsv',
  }


class UnifiedQANatQADPRParaTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for natural_questions_with_dpr_para."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['natural_questions_with_dpr_para']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'natural_questions_with_dpr_para/train.tsv',
      'validation': 'natural_questions_with_dpr_para/dev.tsv',
  }


class UnifiedQANatQADPRParaTestTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for natural_questions_with_dpr_para_test."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['natural_questions_with_dpr_para_test']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'natural_questions_with_dpr_para_test/train.tsv',
      'test': 'natural_questions_with_dpr_para_test/test.tsv',
  }


class UnifiedQANewsQATest(tfds.testing.DatasetBuilderTestCase):
  """Tests for newsqa."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['newsqa']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'newsqa/train.tsv',
      'validation': 'newsqa/dev.tsv',
  }


class UnifiedQAOBQATest(tfds.testing.DatasetBuilderTestCase):
  """Tests for openbookqa."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['openbookqa']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'openbookqa/train.tsv',
      'validation': 'openbookqa/dev.tsv',
      'test': 'openbookqa/test.tsv',
  }


class UnifiedQAOBQADevTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for openbookqa_dev."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['openbookqa_dev']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'openbookqa_dev/train.tsv',
      'validation': 'openbookqa_dev/dev.tsv',
      'test': 'openbookqa_dev/test.tsv',
  }


class UnifiedQAOBQAIRTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for openbookqa_with_ir."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['openbookqa_with_ir']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'openbookqa_with_ir/train.tsv',
      'validation': 'openbookqa_with_ir/dev.tsv',
      'test': 'openbookqa_with_ir/test.tsv',
  }


class UnifiedQAOBQAIRDevTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for openbookqa_with_ir_dev."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['openbookqa_with_ir_dev']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'openbookqa_with_ir_dev/train.tsv',
      'validation': 'openbookqa_with_ir_dev/dev.tsv',
      'test': 'openbookqa_with_ir_dev/test.tsv',
  }


class UnifiedQAPIQATest(tfds.testing.DatasetBuilderTestCase):
  """Tests for physical_iqa."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['physical_iqa']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'physical_iqa/train.tsv',
      'validation': 'physical_iqa/dev.tsv',
  }


class UnifiedQAQASCTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for qasc."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['qasc']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'qasc/train.tsv',
      'validation': 'qasc/dev.tsv',
      'test': 'qasc/test.tsv',
  }


class UnifiedQAQASCTestTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for qasc_test."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['qasc_test']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'qasc_test/train.tsv',
      'validation': 'qasc_test/dev.tsv',
      'test': 'qasc_test/test.tsv',
  }


class UnifiedQAQASCIRTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for qasc_with_ir."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['qasc_with_ir']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'qasc_with_ir/train.tsv',
      'validation': 'qasc_with_ir/dev.tsv',
      'test': 'qasc_with_ir/test.tsv',
  }


class UnifiedQAQASCIRTestTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for qasc_with_ir_test."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['qasc_with_ir_test']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'qasc_with_ir_test/train.tsv',
      'validation': 'qasc_with_ir_test/dev.tsv',
      'test': 'qasc_with_ir_test/test.tsv',
  }


class UnifiedQAQuorefTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for quoref."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['quoref']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'quoref/train.tsv',
      'validation': 'quoref/dev.tsv',
  }


class UnifiedQARACEStringDevTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for race_string_dev."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['race_string_dev']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'race_string_dev/train.tsv',
      'validation': 'race_string_dev/dev.tsv',
      'test': 'race_string_dev/test.tsv',
  }


class UnifiedQARACEStringTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for race_string."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['race_string']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'race_string/train.tsv',
      'validation': 'race_string/dev.tsv',
      'test': 'race_string/test.tsv',
  }


class UnifiedQAROPESTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for ropes."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['ropes']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'ropes/train.tsv',
      'validation': 'ropes/dev.tsv',
  }


class UnifiedQASIQATest(tfds.testing.DatasetBuilderTestCase):
  """Tests for social_iqa."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['social_iqa']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'social_iqa/train.tsv',
      'validation': 'social_iqa/dev.tsv',
  }


class UnifiedQASQuAD11Test(tfds.testing.DatasetBuilderTestCase):
  """Tests for squad1_1."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['squad1_1']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'squad1_1/train.tsv',
      'validation': 'squad1_1/dev.tsv',
  }


class UnifiedQASQuAD20Test(tfds.testing.DatasetBuilderTestCase):
  """Tests for squad2."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['squad2']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'squad2/train.tsv',
      'validation': 'squad2/dev.tsv',
  }


class UnifiedQAWGLTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for winogrande_l."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['winogrande_l']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'winogrande_l/train.tsv',
      'validation': 'winogrande_l/dev.tsv',
  }


class UnifiedQAWGMTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for winogrande_m."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['winogrande_m']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
  }

  DL_EXTRACT_RESULT = {
      'train': 'winogrande_m/train.tsv',
      'validation': 'winogrande_m/dev.tsv',
  }


class UnifiedQAWGSTest(tfds.testing.DatasetBuilderTestCase):
  """Tests for winogrande_s."""

  BUILDER_CONFIG_NAMES_TO_TEST = ['winogrande_s']
  DATASET_CLASS = unifiedqa.UnifiedQA
  SPLITS = {
      'train': 3,  # Number of fake train example
      'validation': 1,  # Number of fake validation example
      'test': 1,  # Number of fake test example
  }

  DL_EXTRACT_RESULT = {
      'train': 'winogrande_s/train.tsv',
      'validation': 'winogrande_s/dev.tsv',
      'test': 'winogrande_s/test.tsv',
  }


if __name__ == '__main__':
  tfds.testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for building a TFDS dataset from RLDS data."""

from __future__ import annotations

import dataclasses
import os
from typing import Any, Dict, Generator, List, Optional, Tuple, Union

from absl import logging
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.rlds import envlogger_reader


@dataclasses.dataclass
class DatasetConfig(tfds.core.BuilderConfig):
  """Configuration of the RLDS dataset.

  Attributes:
    description: string describing the dataset config.
    overall_description: string describing the dataset.
    homepage: URL hosting dataset information.
    citation: Citation for the dataset.
    observation_info: spec of the observation expressed as a TFDS feature.
    action_info: spec of the action expressed as a TFDS feature.
    reward_info: spec of the reward expressed as a TFDS feature.
    discount_info: spec of the discount expressed as a TFDS feature.
    step_metadata_info: spec of the step metadata expresses as a dictionary of
      'feature_name': TFDS feature. Can be an empty dictionary.
    episode_metadata_info: spec of the episode metadata expresses as a
      dictionary of 'feature_name': TFDS feature. Can be an empty dictionary.
    supervised_keys: Supervised keys to be used is `as_supervised=True` in
      `builder.as_dataset`. See `tfds.core.DatasetInfo` for details.
  """

  description: str = ''
  overall_description: str = ''
  homepage: str = ''
  citation: str = ''
  observation_info: Optional[tfds.typing.FeatureConnectorArg] = None
  action_info: Optional[tfds.features.FeatureConnector] = None
  reward_info: Optional[tfds.features.FeatureConnector] = None
  discount_info: Optional[tfds.features.FeatureConnector] = None
  step_metadata_info: Optional[Dict[str, tfds.features.FeatureConnector]] = None
  episode_metadata_info: Optional[Dict[str, tfds.features.FeatureConnector]] = (
      None
  )
  supervised_keys: Optional[Tuple[Any]] = None


def build_info(
    ds_config: DatasetConfig,
    builder: Union[tfds.core.DatasetBuilder, tfds.core.DatasetIdentity],
    ds_metadata: Optional[Dict[Any, Any]] = None,
) -> tfds.core.DatasetInfo:
  """Returns the dataset metadata."""
  step_metadata = ds_config.step_metadata_info
  if step_metadata is None:
    step_metadata = {}
  episode_metadata = ds_config.episode_metadata_info
  if episode_metadata is None:
    episode_metadata = {}
  step_info = {
      'is_terminal': tf.bool,
      'is_first': tf.bool,
      'is_last': tf.bool,
      **step_metadata,
  }
  if ds_config.observation_info:
    step_info['observation'] = ds_config.observation_info
  if ds_config.action_info:
    step_info['action'] = ds_config.action_info
  if ds_config.reward_info:
    step_info['reward'] = ds_config.reward_info
  if ds_config.discount_info:
    step_info['discount'] = ds_config.discount_info
  if ds_metadata:
    ds_metadata = tfds.core.MetadataDict(ds_metadata)
  return tfds.core.DatasetInfo(
      builder=builder,
      description=ds_config.overall_description,
      features=tfds.features.FeaturesDict({
          'steps': tfds.features.Dataset(step_info),
          **episode_metadata,
      }),
      supervised_keys=ds_config.supervised_keys,
      homepage=ds_config.homepage,
      citation=ds_config.citation,
      metadata=ds_metadata,
  )


def get_log_paths(root_dir: str) -> List[str]:
  """Returns the paths of environment logs under a (set of) directories.

  We assume that a sub-directory with metadata.riegeli file contains the logs.

  Args:
    root_dir: Root directory to search for log paths.

  Returns:
    A list of paths that contain the environment logs.

  Raises:
    ValueError if the specified pattern matches a non-directory.
  """
  paths = []
  if not tf.io.gfile.isdir(root_dir):
    raise ValueError(f'{root_dir} is not a directory.')
  for path, _, files in tf.io.gfile.walk(root_dir):
    if 'metadata.riegeli' in files:
      paths.append(path)
  return paths


def generate_examples(path):
  """Yields examples."""
  # TODO(sabela): Consider adding the option of passing a filter to remove
  # some of the episodes.
  tag_dirs = get_log_paths(path.resolve())
  for tag_dir in tag_dirs:
    yield from _generate_examples_from_log_path(tag_dir)


def generate_beam_examples(path):
  """Yields examples using Beam."""
  beam = tfds.core.lazy_imports.apache_beam

  tag_dirs = get_log_paths(path.resolve())
  return beam.Create(tag_dirs) | beam.FlatMap(_generate_examples_from_log_path)


def _generate_examples_from_log_path(
    log_path: str,
) -> Generator[Tuple[str, Dict[str, Any]], None, None]:
  """Yields examples from a directory containing log files.

  Args:
    log_path: Path of the directory containing the log files.

  Yields:
    a unique example ID and dictionary pair for each episode.
  """
  envlogger = tfds.core.lazy_imports.envlogger
  logging.info('Processing directory %s.', log_path)
  counter = 0
  # TFDS suggests using the basename to avoid having the user path in the key.
  key_prefix = os.path.basename(log_path)
  with envlogger.Reader(log_path) as reader:
    for episode_dict in envlogger_reader.generate_episodes(reader):
      # The example ID should be unique.
      episode_id = counter
      if 'episode_id' in episode_dict:
        episode_id = episode_dict['episode_id']
      yield f'{key_prefix}/{episode_id}', episode_dict
      counter += 1



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for feature."""
import numpy as np
import pytest
import tensorflow as tf
from tensorflow_datasets.core.features import feature
from tensorflow_datasets.core.features import image_feature
from tensorflow_datasets.core.proto import feature_pb2


def test_to_shape_proto_single_dimension():
  shape = [1]
  shape_proto = feature.to_shape_proto(shape)
  assert shape_proto.dimensions == [1]


def test_to_shape_proto_single_zero_dimension():
  shape = [0]
  shape_proto = feature.to_shape_proto(shape)
  assert shape_proto.dimensions == [0]


def test_to_shape_proto_normal():
  shape = (28, 28, 1)
  shape_proto = feature.to_shape_proto(shape)
  assert shape_proto.dimensions == [28, 28, 1]


def test_to_shape_proto_unspecified():
  shape = (28, 28, None)
  shape_proto = feature.to_shape_proto(shape)
  assert shape_proto.dimensions == [28, 28, -1]


def test_from_shape_proto_single_dimension():
  shape_proto = feature_pb2.Shape(dimensions=[28])
  assert [28] == feature.from_shape_proto(shape_proto)


def test_from_shape_proto_single_zero_dimension():
  shape_proto = feature_pb2.Shape(dimensions=[0])
  assert [0] == feature.from_shape_proto(shape_proto)


def test_from_shape_proto_normal():
  shape_proto = feature_pb2.Shape(dimensions=[28, 28, 1])
  assert [28, 28, 1] == feature.from_shape_proto(shape_proto)


def test_from_shape_proto_unspecified():
  shape_proto = feature_pb2.Shape(dimensions=[28, 28, -1])
  assert [28, 28, None] == feature.from_shape_proto(shape_proto)


def test_dtype_to_str():
  assert feature.dtype_to_str(np.int64) == "int64"
  assert feature.dtype_to_str(np.uint8) == "uint8"


def test_dtype_from_str():
  assert feature.dtype_from_str("int64") == np.int64
  assert feature.dtype_from_str("uint8") == np.uint8


def test_dtype_for_string_types():
  assert feature.dtype_from_str("str") == np.object_
  assert feature.dtype_from_str("string") == np.object_
  assert feature.dtype_to_str(np.str_) == "string"
  assert feature.dtype_to_str(np.object_) == "string"
  assert feature.dtype_to_str(tf.string) == "string"


def test_encode_and_dtype_from_str():
  dtypes = [
      tf.int64,
      tf.string,
      tf.bfloat16,
      tf.bool,
      np.int64,
      np.bool_,
      np.object_,
      np.uint8,
  ]
  for dtype in dtypes:
    assert feature.dtype_from_str(feature.dtype_to_str(dtype)) == dtype


@pytest.mark.parametrize(
    ["json"],
    [
        (
            {
                "type": "tensorflow_datasets.core.features.image_feature.Image",
                "content": {
                    "shape": [128, 128, 1],
                    "dtype": "uint8",
                    "encoding_format": None,
                    "use_colormap": False,
                },
            },
        ),
        (
            {
                "type": "tensorflow_datasets.core.features.labeled_image.LabeledImage",
                "content": {
                    "shape": [128, 128, 1],
                    "dtype": "uint8",
                    "encoding_format": None,
                    "labels": 2,
                },
            },
        ),
    ],
)
def test_feature_from_json(json):
  feature_connector = feature.FeatureConnector.from_json(json)
  assert isinstance(feature_connector, image_feature.Image)
  assert feature_connector.dtype == np.uint8
  assert feature_connector.shape == (128, 128, 1)


@pytest.mark.parametrize(["dtype"], [(np.int64,), (tf.int64,)])
def test_tensor_info_tensor_shape(dtype):
  tensor_shape = tf.TensorShape([28, 28, 3])
  tensor_info = feature.TensorInfo(shape=tensor_shape, dtype=dtype)
  assert tensor_info.shape == (28, 28, 3)
  assert tensor_info.to_tensor_spec() == tf.TensorSpec(
      shape=tensor_shape, dtype=np.int64
  )


@pytest.mark.parametrize(["dtype"], [(np.int64,), (tf.int64,)])
def test_tensor_info_tensor_shape_with_none(dtype):
  tensor_shape = tf.TensorShape([None, None, 3])
  tensor_info = feature.TensorInfo(shape=tensor_shape, dtype=dtype)
  assert tensor_info.shape == (None, None, 3)
  assert tensor_info.to_tensor_spec() == tf.TensorSpec(
      shape=tensor_shape, dtype=np.int64
  )


@pytest.mark.parametrize(["dtype"], [(np.int64,), (tf.int64,)])
def test_tensor_info_list_shape(dtype):
  tensor_info = feature.TensorInfo(shape=[28, 28, 3], dtype=dtype)
  assert tensor_info.shape == (28, 28, 3)


@pytest.mark.parametrize(["dtype"], [(np.int64,), (tf.int64,)])
def test_tensor_info_list_shape_with_none(dtype):
  tensor_info = feature.TensorInfo(shape=[None, None, 3], dtype=dtype)
  assert tensor_info.shape == (None, None, 3)


@pytest.mark.parametrize(
    ["feature_name", "parent_name", "expected"],
    [
        ("a", None, "a"),
        ("a/b", None, "a.b"),
        ("a", "b", "b-a"),
        ("a/b", "c/d", "c-d-a.b"),
    ],
)
def test_convert_feature_name_to_filename(feature_name, parent_name, expected):
  assert (
      feature.convert_feature_name_to_filename(
          feature_name=feature_name, parent_name=parent_name
      )
      == expected
  )



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Benchmark utils."""
from __future__ import annotations

import dataclasses
import statistics
import textwrap
import time
from typing import Any, Dict, Iterable, List, Optional, Union

from absl import logging
from tensorflow_datasets.core.utils import tqdm_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import pandas as pd

# pylint: disable=logging-format-interpolation

StatDict = Dict[str, Union[int, float]]


def _ns_to_s(time_in_ns: int) -> float:
  return time_in_ns / 1e9


@dataclasses.dataclass(frozen=True)
class RawBenchmarkResult:
  """Raw results of running the benchmark.

  Attributes:
    num_iter: the number of iterations over examples that were done.
    num_examples: the number of examples that were processed in these
      iterations. Note that when examples are batched, then one iteration
      processes multiple examples.
    start_time: the time (in ns) at which when the benchmark started.
    first_batch_time: the time (in ns) at which the first iteration was
      processed.
    end_time: the time (in ns) at which the benchmark ended.
    batch_size: the number of examples in each iteration.
    durations_ns: the duration in ns of each iteration that was processed.
  """

  num_iter: int
  num_examples: int
  start_time: int
  first_batch_time: int
  end_time: int
  batch_size: int
  durations_ns: Optional[List[int]] = None

  def examples(self, include_first: bool = True) -> int:
    if include_first:
      return self.num_examples
    return self.num_examples - 1

  def total_time_s(self, include_first: bool = True) -> float:
    if include_first:
      return _ns_to_s(self.end_time - self.start_time)
    return _ns_to_s(self.end_time - self.first_batch_time)

  def examples_per_second(self, include_first: bool = True) -> float:
    return self.examples(include_first) / self.total_time_s(include_first)

  def time_until_first(self, include_first: bool = True) -> Optional[float]:
    """Time in seconds that it took to load the first example."""
    if include_first:
      return _ns_to_s(self.first_batch_time - self.start_time)
    if self.durations_ns is not None and len(self.durations_ns) > 1:
      return _ns_to_s(self.durations_ns[1])
    return None

  def durations_s(self, include_first: bool = True) -> List[float]:
    if not include_first:
      return [_ns_to_s(d) for d in self.durations_ns[1:]]
    return [_ns_to_s(d) for d in self.durations_ns]

  def summary_statistics(
      self, include_first: bool
  ) -> Dict[str, Union[float, List[float]]]:
    if self.durations_ns is None:
      return {}
    durations = self.durations_s(include_first)
    return {
        'mean': statistics.mean(durations),
        'variance': statistics.pvariance(durations),
        'stdev': statistics.stdev(durations),
        'quantiles': statistics.quantiles(durations),
    }

  def raw_stats_pd(self) -> pd.DataFrame:  # pytype: disable=invalid-annotation  # typed-pandas
    raw_stats = {
        'start_time': _ns_to_s(self.start_time),
        'first_batch_time': _ns_to_s(self.first_batch_time),
        'end_time': _ns_to_s(self.end_time),
        'num_iter': self.num_iter,
    }
    return pd.DataFrame.from_dict(
        raw_stats, orient='index', columns=['duration']
    )

  def stats(self) -> Dict[str, StatDict]:
    return {
        'first+lasts': _log_stats(
            'First included',
            self.start_time,
            self.end_time,
            self.num_examples + self.batch_size,
        ),
        'first': _log_stats(
            'First only',
            self.start_time,
            self.first_batch_time,
            self.batch_size,
        ),
        'lasts': _log_stats(
            'First excluded',
            self.first_batch_time,
            self.end_time,
            self.num_examples,
        ),
    }

  def stats_pd(self) -> pd.DataFrame:  # pytype: disable=invalid-annotation  # typed-pandas
    return pd.DataFrame.from_dict(self.stats(), orient='index')

  def __repr__(self) -> str:
    return textwrap.dedent(
        f"""
      BenchmarkResult(
        num_iter = {self.num_iter}
        num_examples = {self.num_examples}
        batch_size = {self.batch_size}
        examples / sec = {self.examples_per_second()}
        duration: {self.summary_statistics(include_first=True)}
        duration without first: {self.summary_statistics(include_first=False)}
      )"""
    )


@dataclasses.dataclass(frozen=True)
class BenchmarkResult:
  stats: pd.DataFrame  # pytype: disable=invalid-annotation  # typed-pandas
  raw_stats: pd.DataFrame  # pytype: disable=invalid-annotation  # typed-pandas

  def _repr_html_(self) -> str:
    """Colab/notebook representation."""
    return '<strong>BenchmarkResult:</strong><br/>' + self.stats._repr_html_()  # pylint: disable=protected-access


def raw_benchmark(
    ds: Iterable[Any],
    *,
    num_iter: Optional[int] = None,
    batch_size: int = 1,
    detailed_stats: bool = False,
) -> RawBenchmarkResult:
  """Benchmarks any iterable (e.g `tf.data.Dataset`).

  Usage:

  ```py
  ds = tfds.load('mnist', split='train')
  ds = ds.batch(32).prefetch(buffer_size=tf.data.AUTOTUNE)
  tfds.benchmark(ds, batch_size=32)
  ```

  Args:
    ds: Dataset to benchmark. Can be any iterable. Note: The iterable will be
      fully consumed.
    num_iter: Number of iteration to perform (iteration might be batched)
    batch_size: Batch size of the dataset, used to normalize iterations
    detailed_stats: Whether to collect detailed statistics such as the time that
      each iteration took.

  Returns:
    raw results.
  """
  try:
    total = len(ds)  # pytype: disable=wrong-arg-types
  except TypeError:
    total = num_iter

  if num_iter is not None:
    total = min(total, num_iter)

  results = []
  actual_num_iter = 0
  first_batch_time = None
  start_time = time.perf_counter_ns()
  end_time = start_time
  for _ in tqdm_utils.tqdm(iter(ds), total=total):
    actual_num_iter += 1
    end_time = time.perf_counter_ns()
    if first_batch_time is None:
      first_batch_time = end_time
    if detailed_stats:
      results.append(end_time)
    if num_iter and actual_num_iter >= num_iter:
      break
  if not actual_num_iter:
    raise ValueError('Cannot benchmark dataset with 0 elements.')

  if num_iter and actual_num_iter < num_iter:
    logging.warning(
        'Number of iterations is shorter than expected ({} vs {})'.format(
            actual_num_iter, num_iter
        )
    )

  durations_ns = []
  for i in range(len(results)):
    if i == 0:
      durations_ns.append(results[i] - start_time)
    else:
      durations_ns.append(results[i] - results[i - 1])

  return RawBenchmarkResult(
      num_iter=actual_num_iter,
      num_examples=actual_num_iter * batch_size,
      start_time=start_time,
      first_batch_time=first_batch_time,
      end_time=end_time,
      batch_size=batch_size,
      durations_ns=durations_ns,
  )


def benchmark(
    ds: Iterable[Any],
    *,
    num_iter: Optional[int] = None,
    batch_size: int = 1,
) -> BenchmarkResult:
  """Benchmarks any iterable (e.g `tf.data.Dataset`).

  Usage:

  ```py
  ds = tfds.load('mnist', split='train')
  ds = ds.batch(32).prefetch(buffer_size=tf.data.AUTOTUNE)
  tfds.benchmark(ds, batch_size=32)
  ```

  Reports:

  - Total execution time
  - Setup time (first warmup batch)
  - Number of examples/sec

  Args:
    ds: Dataset to benchmark. Can be any iterable. Note: The iterable will be
      fully consumed.
    num_iter: Number of iteration to perform (iteration might be batched)
    batch_size: Batch size of the dataset, used to normalize iterations

  Returns:
    statistics: The recorded statistics, for eventual post-processing
  """
  print('\n************ Summary ************\n')
  raw_results = raw_benchmark(ds=ds, num_iter=num_iter, batch_size=batch_size)
  return BenchmarkResult(
      stats=raw_results.stats_pd(),
      raw_stats=raw_results.raw_stats_pd(),
  )


def _log_stats(msg: str, start: int, end: int, num_examples: int) -> StatDict:
  """Log and returns stats."""
  if not num_examples:
    stats = {
        'duration': 0.0,
        'num_examples': 0,
        'avg': 0.0,
    }
  else:
    # Make sure the total time is not 0.
    total_time_ns = (end - start) or 1
    total_time_s = _ns_to_s(total_time_ns)
    stats = {
        'duration': total_time_s,
        'num_examples': num_examples,
        'avg': num_examples / total_time_s,
    }
  print(
      'Examples/sec ({}) {avg:.2f} ex/sec (total: {num_examples} ex, '
      '{duration:.2f} sec)'.format(msg, **stats)
  )
  return stats



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Access registered datasets."""

import abc
import collections
import contextlib
import functools
import importlib
import inspect
import os.path
from typing import ClassVar, Dict, Iterator, List, Type, Text, Tuple

from etils import epath
from tensorflow_datasets.core import constants
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import visibility
from tensorflow_datasets.core.utils import py_utils
from tensorflow_datasets.core.utils import resource_utils

# Internal registry containing <str registered_name, DatasetBuilder subclass>
_DATASET_REGISTRY = {}

# Internal registry containing:
# <str snake_cased_name, abstract DatasetBuilder subclass>
_ABSTRACT_DATASET_REGISTRY = {}

# Keep track of Dict[str (module name), List[DatasetBuilder]]
# This is directly accessed by `tfds.community.builder_cls_from_module` when
# importing community packages.
_MODULE_TO_DATASETS = collections.defaultdict(list)

# Internal registry containing:
# <str registered_name, DatasetCollectionBuilder subclass>
_DATASET_COLLECTION_REGISTRY = {}

# Internal registry containing:
# <str snake_cased_name, abstract DatasetCollectionBuilder subclass>
_ABSTRACT_DATASET_COLLECTION_REGISTRY = {}

# Keep track of Dict[str (module name), List[DatasetCollectionBuilder]]
_MODULE_TO_DATASET_COLLECTIONS = collections.defaultdict(list)

# eg for dataset "foo": "tensorflow_datasets.datasets.foo.foo_dataset_builder".
_BUILDER_MODULE_SUFFIX = '_dataset_builder'


class DatasetNotFoundError(ValueError):
  """Exception raised when the dataset cannot be found."""


class DatasetCollectionNotFoundError(ValueError):
  """Exception raised when the dataset collection cannot be found."""


_skip_registration = False


@contextlib.contextmanager
def skip_registration() -> Iterator[None]:
  """Context manager within which dataset builders are not registered."""
  global _skip_registration
  try:
    _skip_registration = True
    yield
  finally:
    _skip_registration = False


# The implementation of this class follows closely RegisteredDataset.
class RegisteredDatasetCollection(abc.ABC):
  """Subclasses will be registered and given a `name` property."""

  # Name of the dataset_collection, automatically filled.
  name: ClassVar[str]

  def __init_subclass__(cls, skip_registration=False, **kwargs):  # pylint: disable=redefined-outer-name
    super().__init_subclass__(**kwargs)

    # Set the name if the dataset_collection does not define it.
    # Use __dict__ rather than getattr so subclasses are not affected.
    if not cls.__dict__.get('name'):
      cls.name = naming.camelcase_to_snakecase(cls.__name__)

    is_abstract = inspect.isabstract(cls)

    # Capture all concrete dataset_collections, including when skip registration
    # is True. This ensures that `builder_cls_from_module` can load the
    # dataset_collections even when the module has been imported inside a
    # `skip_registration` context.
    if not is_abstract:
      _MODULE_TO_DATASET_COLLECTIONS[cls.__module__].append(cls)

    # Skip dataset_collection registration within contextmanager, or if
    # skip_registration is passed as meta argument.
    if skip_registration or _skip_registration:
      return

    # Check for name collisions
    if py_utils.is_notebook():  # On Colab/Jupyter, we allow overwriting
      pass
    elif cls.name in _DATASET_COLLECTION_REGISTRY:
      raise ValueError(
          f'DatasetCollection with name {cls.name} already registered.'
      )
    elif cls.name in _ABSTRACT_DATASET_COLLECTION_REGISTRY:
      raise ValueError(
          f'DatasetCollection with name {cls.name} already registered as'
          ' abstract.'
      )

    # Add the dataset_collection to the registers
    if is_abstract:
      _ABSTRACT_DATASET_COLLECTION_REGISTRY[cls.name] = cls
    else:
      _DATASET_COLLECTION_REGISTRY[cls.name] = cls


def list_imported_dataset_collections() -> List[str]:
  """Returns the string names of all `tfds.core.DatasetCollection`s."""
  all_dataset_collections = [
      dataset_collection_name
      for dataset_collection_name, dataset_collection_cls in _DATASET_COLLECTION_REGISTRY.items()
  ]
  return sorted(all_dataset_collections)


def is_dataset_collection(name: str) -> bool:
  return name in _DATASET_COLLECTION_REGISTRY


def imported_dataset_collection_cls(
    name: str,
) -> Type[RegisteredDatasetCollection]:
  """Returns the Registered dataset class."""
  if name in _ABSTRACT_DATASET_COLLECTION_REGISTRY:
    raise AssertionError(f'DatasetCollection {name} is an abstract class.')

  if not is_dataset_collection(name):
    raise DatasetCollectionNotFoundError(f'DatasetCollection {name} not found.')

  dataset_collection_cls = _DATASET_COLLECTION_REGISTRY[name]

  return dataset_collection_cls  # pytype: disable=bad-return-type


class RegisteredDataset(abc.ABC):
  """Subclasses will be registered and given a `name` property."""

  # Name of the dataset, automatically filled if not already set.
  name: ClassVar[str]


  def __init_subclass__(cls, skip_registration=False, **kwargs):  # pylint: disable=redefined-outer-name
    super().__init_subclass__(**kwargs)

    # Set the name if the dataset does not define it.
    # Use __dict__ rather than getattr so subclasses are not affected.
    if not cls.__dict__.get('name'):
      if cls.__name__ == 'Builder':
        # Config-based builders should be defined with a class named "Builder".
        # In such a case, the builder name is extracted from the module if it
        # follows conventions:
        module_name = cls.__module__.rsplit('.', 1)[-1]
        if module_name.endswith(_BUILDER_MODULE_SUFFIX):
          cls.name = module_name[: -len(_BUILDER_MODULE_SUFFIX)]
        elif '.' in cls.__module__:  # Extract dataset name from package name.
          cls.name = cls.__module__.rsplit('.', 2)[-2]
        else:
          raise AssertionError(
              'When using `Builder` as class name, the dataset builder name is '
              'inferred from module name if named "*_dataset_builder" or from '
              f'package name, but there is no package in "{cls.__module__}".'
          )
      else:  # Legacy builders.
        cls.name = naming.camelcase_to_snakecase(cls.__name__)

    is_abstract = inspect.isabstract(cls)

    # Capture all concrete datasets, including when skip registration is True.
    # This ensures that `builder_cls_from_module` can load the datasets
    # even when the module has been imported inside a `skip_registration`
    # context.
    if not is_abstract:
      _MODULE_TO_DATASETS[cls.__module__].append(cls)

    # Skip dataset registration within contextmanager, or if skip_registration
    # is passed as meta argument.
    if skip_registration or _skip_registration:
      return

    # Check for name collisions
    if py_utils.is_notebook():  # On Colab/Jupyter, we allow overwriting
      pass
    elif cls.name in _DATASET_REGISTRY:
      raise ValueError(f'Dataset with name {cls.name} already registered.')
    elif cls.name in _ABSTRACT_DATASET_REGISTRY:
      raise ValueError(
          f'Dataset with name {cls.name} already registered as abstract.'
      )

    # Add the dataset to the registers
    if is_abstract:
      _ABSTRACT_DATASET_REGISTRY[cls.name] = cls
    else:
      _DATASET_REGISTRY[cls.name] = cls


def _is_builder_available(builder_cls: Type[RegisteredDataset]) -> bool:
  """Returns `True` is the builder is available."""
  return visibility.DatasetType.TFDS_PUBLIC.is_available()


def list_imported_builders() -> List[str]:
  """Returns the string names of all `tfds.core.DatasetBuilder`s."""
  all_builders = [
      builder_name
      for builder_name, builder_cls in _DATASET_REGISTRY.items()
      if _is_builder_available(builder_cls)
  ] + list(_get_existing_dataset_packages(constants.DATASETS_TFDS_SRC_DIR))
  return sorted(all_builders)


@functools.lru_cache(maxsize=None)
def _get_existing_dataset_packages(
    datasets_dir: Text,
) -> Dict[Text, Tuple[epath.Path, Text]]:
  """Returns existing datasets.

  Args:
    datasets_dir: directory path, relative to tensorflow_datasets srcs root,
      where are defined the dataset packages. This directory must only contain
      valid dataset packages.

  Returns:
    {ds_name: (pkg_path, builder_module)}.
    For example: {'mnist': ('/lib/tensorflow_datasets/datasets/mnist',
                            'tensorflow_datasets.datasets.mnist.builder')}
  """
  datasets = {}
  try:
    datasets_dir_path = resource_utils.tfds_path(datasets_dir)
  except OSError:
    # Raised when datasets_dir does not exist, for example in tests when data
    # does not contain the directory (when running with bazel).
    return datasets
  if not datasets_dir_path.exists():
    return datasets
  ds_dir_pkg = '.'.join(
      ['tensorflow_datasets'] + datasets_dir.split(os.path.sep)
  )
  for child in datasets_dir_path.iterdir():
    # Except for a few exceptions, all children of datasets/ directory are
    # packages of datasets, no needs to check child is a directory and contains
    # a `builder.py` module.
    exceptions = [
        '__init__.py',
    ]
    if child.name not in exceptions:
      pkg_path = epath.Path(datasets_dir_path) / child.name
      builder_module = (
          f'{ds_dir_pkg}.{child.name}.{child.name}{_BUILDER_MODULE_SUFFIX}'
      )
      datasets[child.name] = (pkg_path, builder_module)
  return datasets


def imported_builder_cls(name: str) -> Type[RegisteredDataset]:
  """Returns the Registered dataset class."""
  existing_ds_pkgs = _get_existing_dataset_packages(
      constants.DATASETS_TFDS_SRC_DIR
  )
  if name in existing_ds_pkgs:
    pkg_dir_path, builder_module = existing_ds_pkgs[name]
    cls = importlib.import_module(builder_module).Builder
    cls.pkg_dir_path = pkg_dir_path
    return cls

  if name in _ABSTRACT_DATASET_REGISTRY:
    # Will raise TypeError: Can't instantiate abstract class X with abstract
    # methods y, before __init__ even get called
    _ABSTRACT_DATASET_REGISTRY[name]()  # pytype: disable=not-callable
    # Alternatively, could manually extract the list of non-implemented
    # abstract methods.
    raise AssertionError(f'Dataset {name} is an abstract class.')

  if name not in _DATASET_REGISTRY:
    raise DatasetNotFoundError(f'Dataset {name} not found.')

  builder_cls = _DATASET_REGISTRY[name]
  if not _is_builder_available(builder_cls):
    available_types = visibility.get_availables()
    msg = f'Dataset {name} is not available. Only: {available_types}'
    raise PermissionError(msg)
  return builder_cls  # pytype: disable=bad-return-type



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A parser for ranking-style LibSVM files.

Note that the LibSVM ranking file format does not have a formal specification.
This implementation supports the common formats that are available online and
will work with all the publically available LTR datasets such as MSLR-WEB,
Istella and Yahoo Webscope.
"""

import collections
import dataclasses
import re
from typing import Iterable, List, Mapping, Tuple

import numpy as np

# Type alias for a tuple representing a ranking example.
RankingExampleTuple = Tuple[str, Mapping[str, np.ndarray]]


@dataclasses.dataclass
class RankingExample:
  """Represents a parsed ranking example for a given query identifier.

  Attributes:
    qid: The query identifier.
    features: A mapping of feature name to feature values.
  """

  qid: str
  features: Mapping[str, List[float]]


class ParserError(Exception):
  """Raised when LibSVM-formatted contents cannot be parsed correctly."""

  def __init__(self, line_number: int, line: str, reason: str):
    """Initializes the instance.

    Args:
      line_number: The line number where the parser error occurred.
      line: The line where the parser error occurred.
      reason: An informative message about the nature of the parser error.
    """
    super().__init__(f"Unable to parse line {line_number} ('{line}'): {reason}")
    self.line_number = line_number
    self.line = line
    self.reason = reason


class LibSVMRankingParser(Iterable[RankingExampleTuple]):
  """A parser for LibSVM-formatted ranking files.

  This parser can parse an iterable of lines into an iterable of
  `RankingExampleTuple`s that contain the per-query labels and features.

  Example usage:
  >>> lines = ["1 qid:1 1:5.0 2:3.0", "2 qid:1 1:2.0 2:1.0"]
  >>> feature_names = {1: "bm25", 2: "tfidf"}
  >>> parser = LibSVMRankingParser(lines, feature_names)
  >>> for example in parser:
  ...   print(example[0])
  {'bm25': array([5., 2.]), 'tfidf': array([3., 1.]), 'label': array([1, 2])}

  The parser will raise a ParserError with an informative error message if the
  content being parsed does not match LibSVM ranking format.
  """

  def __init__(
      self,
      lines: Iterable[str],
      feature_names: Mapping[int, str],
      label_feature_name: str = "label",
      query_id_feature_name: str = "query_id",
      doc_id_feature_name: str = "doc_id",
      default_feature_value: float = 0.0,
      combine_features: bool = False,
  ):
    """Initializes the instance.

    Args:
      lines: The lines to parse.
      feature_names: A mapping from feature indices to feature names.
      label_feature_name: The name to assign to the label feature.
      query_id_feature_name: The name to assign to the query identifier feature.
      doc_id_feature_name: The name to assign to the document identifier
        feature.
      default_feature_value: The default feature value to use when a feature is
        missing from the input.
      combine_features: Whether to combine the features into a single
        'float_features' feature.
    """
    self._lines = lines
    self._feature_names = feature_names
    self._label_feature_name = label_feature_name
    self._query_id_feature_name = query_id_feature_name
    self._doc_id_feature_name = doc_id_feature_name
    self._default_feature_value = default_feature_value
    self._current_example = None
    self._available_examples = collections.deque()
    self._combine_features = combine_features

  def _parse_line(self, line_number: int, line: str):
    """Parses a single line of input from a LibSVM ranking file.

    This method will update the internal state of the parser and may produce a
    new item to be made available in `self._available_examples`.

    Args:
      line_number: The current line number. This may be used to generate more
        informative ParserError messages if such an error occurs.
      line: The line to parse.

    Raises:
      ParserError: If the parsing of given line failed.
    """
    # Remove comments and leading/trailing whitespace from line.
    line_clean, *_ = line.split("#", maxsplit=2)
    line_clean = line_clean.strip()

    # An empty line is allowed but should be skipped during parsing.
    if not line_clean:
      return

    # Split by space separators.
    try:
      label, qid, *features = re.split(r"\s+", line_clean)
    except ValueError as value_error:
      raise ParserError(
          line_number, line, "could not extract label, qid and features"
      ) from value_error

    # Convert relevance label to float.
    try:
      label = float(label)
    except ValueError as value_error:
      raise ParserError(
          line_number,
          line,
          f"label '{label}' could not be converted to a float",
      ) from value_error

    # Extract qid.
    if qid[:4] != "qid:":
      raise ParserError(
          line_number, line, "line must contain a qid after the relevance label"
      )
    qid = qid[4:]
    if not qid:
      raise ParserError(line_number, line, "qid can not be empty")

    # Construct a feature dict containing default values.
    feature_dict = {
        feature_name: self._default_feature_value
        for feature_name in self._feature_names.values()
    }

    # Parse all features and add them to the feature dict.
    for feature in features:
      try:
        index, value = feature.split(":", maxsplit=2)
        index = int(index)
        value = float(value)
        # Only add features if they map to a feature name in the
        # `_feature_names` dict. All other features are ignored.
        if index in self._feature_names:
          feature_dict[self._feature_names[index]] = value
      except ValueError as value_error:
        raise ParserError(
            line_number,
            line,
            f"failed to extract feature index and value from '{feature}'",
        ) from value_error

    # Add label to feature dict.
    feature_dict[self._label_feature_name] = label

    # Add document identifier. We use the line number as a unique document
    # identifier since LibSVMRanking files do not have any other way to identify
    # documents.
    feature_dict[self._doc_id_feature_name] = line_number

    # Add the parsed qid and feature dictionary to the current example.
    self._add_to_current_example(qid, feature_dict)

  def _add_to_current_example(self, qid: str, feature_map: Mapping[str, float]):
    """Adds given qid and feature_map to the current example.

    If the qid matches the current example qid, this will add the features in
    given feature_map to the current example.
    If the qid does not match the current example qid, this will store the
    current example and then create a new one with given qid and feature_map.

    Args:
      qid: The query identifier.
      feature_map: A mapping of feature names to feature values.
    """
    if self._current_example is None:
      self._current_example = RankingExample(qid, collections.defaultdict(list))

    if self._current_example.qid != qid:
      self._store_current_example()
      self._current_example = RankingExample(qid, collections.defaultdict(list))

    for key, value in feature_map.items():
      self._current_example.features[key].append(value)

  def _store_current_example(self):
    """Store the current ranking example in numpy format.

    This method converts the current example to a RankingExampleTuple and stores
    it in `self._available_examples` so that it can be yielded by
    `self.__iter__`.
    """
    if not self._current_example:
      return
    qid = self._current_example.qid
    np_features_dict = {
        key: np.array(value)
        for key, value in self._current_example.features.items()
    }

    # Extracts the individual features, sorted by their numerical indices, and
    # re-adds them as a single feature with shape (list_size, feature_size).
    if self._combine_features:
      features = [
          np_features_dict.pop(self._feature_names[idx])
          for idx in sorted(self._feature_names)
      ]
      np_features_dict["float_features"] = np.stack(features, axis=-1)

    # Add query identifier.
    np_features_dict[self._query_id_feature_name] = qid

    self._available_examples.append((qid, np_features_dict))

  def _end_of_parse(self):
    """Signals the end of parsing has been reached.

    This method will store the item that is currently being parsed so it is
    available in `self._available_examples`.
    """
    if self._current_example is not None:
      self._store_current_example()
      self._current_example = None

  def __iter__(self):
    """Iterates over `RankingExampleTuple`s.

    Yields:
      `RankingExampleTuple` that represent a single query.

    Raises:
      ParserError: If parsing failed.
    """
    # Parse each line and yield the resulting examples as they become available.
    for line_number, line in enumerate(self._lines, start=1):
      self._parse_line(line_number, line)
      while self._available_examples:
        yield self._available_examples.popleft()

    # Signal the end has been reached and yield any remaining items.
    self._end_of_parse()
    while self._available_examples:
      yield self._available_examples.popleft()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Script which generates datasets figures.
"""

import functools
import os
import tempfile

from absl import flags
import matplotlib
import matplotlib.pyplot as plt
import tensorflow_datasets as tfds
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf
from tensorflow_datasets.scripts.documentation import script_utils

FLAGS = flags.FLAGS

flags.DEFINE_string(
    'datasets',
    None,
    'Comma separated list of datasets to generates. None for all datasets.',
)
flags.DEFINE_string('dst_dir', None, 'Destination dir to save the images.')
flags.DEFINE_boolean(
    'overwrite', False, 'If True, overwrite the existing visualizations.'
)


def _save_fig(dst_path: str, figure: matplotlib.figure.Figure) -> None:
  """Save the generated figures for the dataset in dst_dir."""
  # `savefig` do not support GCS, so first save the image locally.
  with tempfile.TemporaryDirectory() as tmp_dir:
    tmp_path = os.path.join(tmp_dir, 'tmp.png')
    figure.savefig(tmp_path)
    tf.io.gfile.copy(tmp_path, dst_path, overwrite=FLAGS.overwrite)
  plt.close(figure)


def main(_):
  """Main script."""
  datasets = FLAGS.datasets.split(',') if FLAGS.datasets else None
  generate_and_save_figure_fn = functools.partial(
      script_utils.generate_and_save_artifact,
      dst_dir=FLAGS.dst_dir or tfds.core.gcs_path('visualization/fig'),
      overwrite=FLAGS.overwrite,
      file_extension='.png',
      get_artifact_fn=tfds.show_examples,
      save_artifact_fn=_save_fig,
  )
  script_utils.multi_process_map(
      worker_fn=generate_and_save_figure_fn,
      datasets=datasets,
  )


if __name__ == '__main__':
  script_utils.multi_process_run(main)



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""WMT19: Translate dataset."""

import tensorflow_datasets.public_api as tfds
from tensorflow_datasets.translate import wmt

_URL = "http://www.statmt.org/wmt19/translation-task.html"
# TODO(adarob): Update with citation of overview paper once it is published.
_CITATION = """
@ONLINE {wmt19translate,
    author = "Wikimedia Foundation",
    title  = "ACL 2019 Fourth Conference on Machine Translation (WMT19), Shared Task: Machine Translation of News",
    url    = "http://www.statmt.org/wmt19/translation-task.html"
}
"""

_LANGUAGE_PAIRS = [
    (lang, "en") for lang in ["cs", "de", "fi", "gu", "kk", "lt", "ru", "zh"]
] + [("fr", "de")]


class Wmt19Translate(wmt.WmtTranslate):
  """WMT 19 translation datasets for {(xx, "en")} + ("fr", "de") pairs."""

  # Version history:
  # 1.0.0: S3 (new shuffling, sharding and slicing mechanism).
  BUILDER_CONFIGS = [
      wmt.WmtConfig(  # pylint:disable=g-complex-comprehension
          description="WMT 2019 %s-%s translation task dataset." % (l1, l2),
          url=_URL,
          citation=_CITATION,
          language_pair=(l1, l2),
          version=tfds.core.Version("1.0.0"),
      )
      for l1, l2 in _LANGUAGE_PAIRS
  ]

  @property
  def _subsets(self):
    return {
        tfds.Split.TRAIN: [
            "europarl_v9",
            "europarl_v7_frde",
            "paracrawl_v3",
            "paracrawl_v1_ru",
            "paracrawl_v3_frde",
            "commoncrawl",
            "commoncrawl_frde",
            "newscommentary_v14",
            "newscommentary_v14_frde",
            "czeng_17",
            "yandexcorpus",
            "wikititles_v1",
            "uncorpus_v1",
            "rapid_2016_ltfi",
            "rapid_2019",
        ] + wmt.CWMT_SUBSET_NAMES,
        tfds.Split.VALIDATION: [
            "euelections_dev2019",
            "newsdev2019",
            "newstest2018",
        ],
    }



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for tensorflow_datasets.core.dataset_info."""

import json
import os
import pathlib
import tempfile
import time
from typing import Union

import numpy as np
import pytest
import tensorflow as tf
from tensorflow_datasets import testing
from tensorflow_datasets.core import dataset_info
from tensorflow_datasets.core import features
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core import naming
from tensorflow_datasets.core import read_only_builder
from tensorflow_datasets.core import splits as splits_lib
from tensorflow_datasets.core import utils
from tensorflow_datasets.core.proto import dataset_info_pb2
from tensorflow_datasets.core.proto import feature_pb2
from tensorflow_datasets.image_classification import mnist
from tensorflow_datasets.testing.dummy_config_based_datasets.dummy_ds_1 import dummy_ds_1_dataset_builder

from google.protobuf import text_format


_TFDS_DIR = utils.tfds_path()
_DATA_DIR = os.path.join(_TFDS_DIR, "testing", "test_data", "dataset_info")
_INFO_DIR = os.path.join(_DATA_DIR, "mnist", "3.0.1")
_INFO_DIR_UNLABELED = os.path.join(_DATA_DIR, "mnist_unlabeled", "3.0.1")
_NON_EXISTENT_DIR = os.path.join(_TFDS_DIR, "non_existent_dir")

DummyDatasetSharedGenerator = testing.DummyDatasetSharedGenerator


class RandomShapedImageGenerator(DummyDatasetSharedGenerator):

  def _info(self):
    return dataset_info.DatasetInfo(
        builder=self,
        features=features.FeaturesDict({"im": features.Image()}),
        supervised_keys=("im", "im"),
        metadata=dataset_info.MetadataDict(),
    )

  def _generate_examples(self, range_):
    self.info.metadata["some_key"] = 123

    for i in range_:
      height = np.random.randint(5, high=10)
      width = np.random.randint(5, high=10)
      yield i, {
          "im": np.random.randint(
              0, 255, size=(height, width, 3), dtype=np.uint8
          )
      }


class DatasetInfoTest(testing.TestCase):

  @classmethod
  def setUpClass(cls):
    super(DatasetInfoTest, cls).setUpClass()
    cls._tfds_tmp_dir = testing.make_tmp_dir()
    cls._builder = mnist.MNIST(data_dir=cls._tfds_tmp_dir)

  @classmethod
  def tearDownClass(cls):
    super(DatasetInfoTest, cls).tearDownClass()
    testing.rm_tmp_dir(cls._tfds_tmp_dir)

  def test_non_existent_dir(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    with self.assertRaisesWithPredicateMatch(
        FileNotFoundError, "from a directory which does not exist"
    ):
      info.read_from_directory(_NON_EXISTENT_DIR)

  def test_reading_different_version(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    info._identity.version = utils.Version("2.0.0")
    with pytest.raises(
        AssertionError,
        match=(
            "The constructed DatasetInfo instance and the restored proto"
            " version do not match"
        ),
    ):
      # The dataset in _INFO_DIR has version 3.0.1 whereas the builder is 2.0.0
      info.read_from_directory(_INFO_DIR)

  def test_reading(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    info.read_from_directory(_INFO_DIR)

    # Assert that we read the file and initialized DatasetInfo.
    self.assertTrue(info.initialized)
    self.assertEqual("mnist", info.name)
    self.assertEqual("mnist/3.0.1", info.full_name)

    # Test splits are initialized properly.
    split_dict = info.splits

    # Assert they are the correct number.
    self.assertTrue(len(split_dict), 2)

    # Assert on what they are
    self.assertIn("train", split_dict)
    self.assertIn("test", split_dict)

    # Assert that this is computed correctly.
    self.assertEqual(40, info.splits.total_num_examples)
    self.assertEqual(11594722, info.dataset_size)

    self.assertEqual("image", info.supervised_keys[0])
    self.assertEqual("label", info.supervised_keys[1])
    self.assertEqual(
        info.module_name, "tensorflow_datasets.image_classification.mnist"
    )
    self.assertEqual(False, info.disable_shuffling)

    self.assertEqual(info.version, utils.Version("3.0.1"))
    self.assertEqual(info.release_notes, {})

  def test_disable_shuffling(self):
    info = dataset_info.DatasetInfo(
        builder=self._builder, disable_shuffling=True
    )
    info.read_from_directory(_INFO_DIR)

    self.assertEqual(True, info.disable_shuffling)

  def test_reading_empty_properties(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    info.read_from_directory(_INFO_DIR_UNLABELED)

    # Assert supervised_keys has not been set
    self.assertIsNone(None, info.supervised_keys)

  def test_writing(self):
    # First read in stuff.
    mnist_builder = mnist.MNIST(data_dir=_DATA_DIR)

    info = dataset_info.DatasetInfo(
        builder=mnist_builder, features=mnist_builder.info.features
    )
    info.read_from_directory(_INFO_DIR)

    # Read the json file into a string.
    with tf.io.gfile.GFile(dataset_info.dataset_info_path(_INFO_DIR)) as f:
      existing_json = json.load(f)

    # Now write to a temp directory.
    with testing.tmp_dir(self.get_temp_dir()) as tmp_dir:
      info.write_to_directory(tmp_dir)

      # Read the newly written json file into a string.
      with tf.io.gfile.GFile(dataset_info.dataset_info_path(tmp_dir)) as f:
        new_json = json.load(f)

      # Read the newly written LICENSE file into a string.
      with tf.io.gfile.GFile(dataset_info.license_path(tmp_dir)) as f:
        license_ = f.read()

    # Assert what was read and then written and read again is the same.
    self.assertEqual(existing_json, new_json)

    # Assert correct license was written.
    self.assertEqual(existing_json["redistributionInfo"]["license"], license_)

    # Do not check the full string as it display the generated path.
    self.assertEqual(_INFO_STR % mnist_builder.data_dir, repr(info))
    self.assertIn("'test': <SplitInfo num_examples=", repr(info))

  def test_restore_after_modification(self):
    # Create a DatasetInfo
    info = dataset_info.DatasetInfo(
        builder=self._builder,
        description="A description",
        supervised_keys=("input", "output"),
        homepage="http://some-location",
        citation="some citation",
        license="some license",
    )
    info.as_proto.config_tags.extend(["foo", "bar"])
    info.download_size = 456
    filepath_template = "{DATASET}-{SPLIT}.{FILEFORMAT}-{SHARD_X_OF_Y}"
    info.as_proto.splits.add(
        name="train", num_bytes=512, filepath_template=filepath_template
    )
    info.as_proto.splits.add(
        name="validation", num_bytes=64, filepath_template=filepath_template
    )
    info.as_proto.schema.feature.add()
    info.as_proto.schema.feature.add()  # Add dynamic statistics
    info.download_checksums = {
        "url1": "some checksum",
        "url2": "some other checksum",
    }

    with testing.tmp_dir(self.get_temp_dir()) as tmp_dir:
      # Save it
      info.write_to_directory(tmp_dir)

      # If fields are not defined, then everything is restored from disk.
      restored_info = dataset_info.DatasetInfo(builder=self._builder)
      restored_info.read_from_directory(tmp_dir)
      self.assertProtoEquals(info.as_proto, restored_info.as_proto)

    with testing.tmp_dir(self.get_temp_dir()) as tmp_dir:
      # Save it
      info.write_to_directory(tmp_dir)

      # If fields are defined, then the code version is kept
      restored_info = dataset_info.DatasetInfo(
          builder=self._builder,
          supervised_keys=("input (new)", "output (new)"),
          homepage="http://some-location-new",
          citation="some citation (new)",
          redistribution_info={"license": "some license (new)"},
      )
      restored_info.download_size = 789
      restored_info.as_proto.splits.add(name="validation", num_bytes=288)
      restored_info.as_proto.schema.feature.add()
      restored_info.as_proto.schema.feature.add()
      restored_info.as_proto.schema.feature.add()
      restored_info.as_proto.schema.feature.add()  # Add dynamic statistics
      restored_info.download_checksums = {
          "url2": "some other checksum (new)",
          "url3": "some checksum (new)",
      }

      restored_info.read_from_directory(tmp_dir)

      # Even though restored_info has been restored, informations defined in
      # the code overwrite informations from the json file.
      self.assertEqual(restored_info.description, "A description")
      self.assertEqual(restored_info.version, utils.Version("3.0.1"))
      self.assertEqual(restored_info.release_notes, {})
      self.assertEqual(
          restored_info.supervised_keys, ("input (new)", "output (new)")
      )
      self.assertEqual(restored_info.homepage, "http://some-location-new")
      self.assertEqual(restored_info.citation, "some citation (new)")
      self.assertEqual(
          restored_info.redistribution_info.license, "some license (new)"
      )
      self.assertEqual(restored_info.download_size, 789)
      self.assertEqual(restored_info.dataset_size, 576)
      self.assertEqual(len(restored_info.as_proto.schema.feature), 4)
      self.assertEqual(
          restored_info.download_checksums,
          {
              "url2": "some other checksum (new)",
              "url3": "some checksum (new)",
          },
      )

  def test_reading_from_gcs_bucket(self):
    # The base TestCase prevents GCS access, so we explicitly ask it to restore
    # access here.
    with self.gcs_access():
      mnist_builder = mnist.MNIST(
          data_dir=tempfile.mkdtemp(dir=self.get_temp_dir())
      )
      info = dataset_info.DatasetInfo(builder=mnist_builder)
      info = mnist_builder.info

      # A nominal check to see if we read it.
      self.assertTrue(info.initialized)
      self.assertEqual(10000, info.splits["test"].num_examples)

  def test_str_smoke(self):
    info = mnist.MNIST(data_dir="/tmp/some_dummy_dir").info
    _ = str(info)

  def test_metadata(self):
    with testing.tmp_dir(self.get_temp_dir()) as tmp_dir:
      builder = RandomShapedImageGenerator(data_dir=tmp_dir)
      builder.download_and_prepare()
      # Metadata should have been created
      self.assertEqual(builder.info.metadata, {"some_key": 123})

      # Metadata should have been restored
      builder2 = RandomShapedImageGenerator(data_dir=tmp_dir)
      self.assertEqual(builder2.info.metadata, {"some_key": 123})

      # Metadata should have been restored even if the builder code was not
      # available and we restored from files.
      builder3 = read_only_builder.builder_from_files(
          builder.name,
          data_dir=tmp_dir,
      )
      self.assertEqual(builder3.info.metadata, {"some_key": 123})

  def test_redistribution_info(self):
    info = dataset_info.DatasetInfo(
        builder=self._builder, license="some license"
    )
    info_proto = info.as_proto
    self.assertEqual(info_proto.redistribution_info.license, "some license")

  def test_updates_on_bucket_info(self):
    info = dataset_info.DatasetInfo(
        builder=self._builder, description="won't be updated"
    )
    # No statistics in the above.
    self.assertEqual(0, info.splits.total_num_examples)
    self.assertEqual(0, len(info.as_proto.schema.feature))

    # Partial update will happen here.
    info.read_from_directory(_INFO_DIR)

    # Assert that description (things specified in the code) didn't change
    # but statistics are updated.
    self.assertEqual("won't be updated", info.description)

    # These are dynamically computed, so will be updated.
    self.assertEqual(40, info.splits.total_num_examples)
    self.assertEqual(2, len(info.as_proto.schema.feature))

  def test_set_splits_normal(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    split_info1 = splits_lib.SplitInfo(
        name="train", shard_lengths=[1, 2], num_bytes=0
    )
    split_info2 = splits_lib.SplitInfo(
        name="test", shard_lengths=[1], num_bytes=0
    )
    split_dict = splits_lib.SplitDict(split_infos=[split_info1, split_info2])
    info.set_splits(split_dict)
    self.assertEqual(str(info.splits), str(split_dict))
    self.assertEqual(
        str(info.as_proto.splits),
        str([split_info1.to_proto(), split_info2.to_proto()]),
    )

  def test_set_splits_incorrect_dataset_name(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    split_info1 = splits_lib.SplitInfo(
        name="train",
        shard_lengths=[1, 2],
        num_bytes=0,
        filename_template=naming.ShardedFileTemplate(
            dataset_name="some_other_dataset",
            split="train",
            data_dir=info.data_dir,
            filetype_suffix="tfrecord",
        ),
    )
    split_dict = splits_lib.SplitDict(split_infos=[split_info1])
    with pytest.raises(
        AssertionError, match="SplitDict contains SplitInfo for split"
    ):
      info.set_splits(split_dict)

  def test_set_splits_multi_split_info(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    split_info1 = splits_lib.SplitInfo(
        name="train", shard_lengths=[1, 2], num_bytes=0
    )
    split_info2 = splits_lib.SplitInfo(
        name="test", shard_lengths=[1], num_bytes=0
    )
    multi_split_info1 = splits_lib.MultiSplitInfo(
        name="train", split_infos=[split_info1]
    )
    multi_split_info2 = splits_lib.MultiSplitInfo(
        name="test", split_infos=[split_info2]
    )
    split_dict = splits_lib.SplitDict(
        split_infos=[multi_split_info1, multi_split_info2]
    )
    info.set_splits(split_dict)
    self.assertEqual(str(info.splits), str(split_dict))
    self.assertEqual(
        str(info.as_proto.splits),
        str([split_info1.to_proto(), split_info2.to_proto()]),
    )

  def test_set_file_format_override_fails(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    info.set_file_format(file_adapters.FileFormat.TFRECORD)
    self.assertEqual(info.file_format, file_adapters.FileFormat.TFRECORD)
    with pytest.raises(
        ValueError,
        match=(
            "File format is already set to FileFormat.TFRECORD. Got"
            " FileFormat.RIEGELI"
        ),
    ):
      info.set_file_format(file_adapters.FileFormat.RIEGELI)

  def test_set_file_format_override(self):
    info = dataset_info.DatasetInfo(builder=self._builder)
    info.set_file_format(file_adapters.FileFormat.TFRECORD)
    self.assertEqual(info.file_format, file_adapters.FileFormat.TFRECORD)
    info.set_file_format(file_adapters.FileFormat.RIEGELI, override=True)
    self.assertEqual(info.file_format, file_adapters.FileFormat.RIEGELI)


@pytest.mark.parametrize(
    "file_format",
    [
        file_adapters.FileFormat.TFRECORD,
    ],
)
def test_file_format_save_restore(
    tmp_path: pathlib.Path,
    file_format: file_adapters.FileFormat,
):
  builder = testing.DummyDataset(data_dir=tmp_path, file_format=file_format)

  assert isinstance(builder.info.file_format, file_adapters.FileFormat)
  assert builder.info.file_format is file_format

  builder.download_and_prepare()

  # When restoring the builder, we do not provide the `file_format=`
  # yet it is correctly restored
  builder2 = testing.DummyDataset(data_dir=tmp_path)
  assert builder2.info.file_format is file_format

  # Explicitly passing the correct format is accepted.
  builder3 = testing.DummyDataset(data_dir=tmp_path, file_format=file_format)
  assert builder3.info.file_format is file_format

  # Providing an inconsistent format is rejected.
  with pytest.raises(ValueError, match="File format is already set to"):
    different_file_format = {
        file_adapters.FileFormat.TFRECORD: file_adapters.FileFormat.RIEGELI,
        file_adapters.FileFormat.RIEGELI: file_adapters.FileFormat.TFRECORD,
    }[file_format]
    testing.DummyDataset(data_dir=tmp_path, file_format=different_file_format)


def test_file_format_values(tmp_path: pathlib.Path):
  # Default file format
  builder = testing.DummyDataset(data_dir=tmp_path, file_format=None)
  assert builder.info.file_format == file_adapters.FileFormat.TFRECORD

  # str accepted
  builder = testing.DummyDataset(data_dir=tmp_path, file_format="riegeli")
  assert builder.info.file_format == file_adapters.FileFormat.RIEGELI

  # file_adapters.FileFormat accepted
  builder = testing.DummyDataset(
      data_dir=tmp_path, file_format=file_adapters.FileFormat.RIEGELI
  )
  assert builder.info.file_format == file_adapters.FileFormat.RIEGELI

  # Unknown value
  with pytest.raises(ValueError, match="is not a valid FileFormat"):
    testing.DummyDataset(data_dir=tmp_path, file_format="arrow")


@pytest.fixture(scope="session")
def generator_builder():
  return RandomShapedImageGenerator(data_dir=testing.make_tmp_dir())


def test_dataset_info_from_proto(generator_builder):  # pylint: disable=redefined-outer-name
  train = dataset_info_pb2.SplitInfo(
      name="train", num_shards=2, shard_lengths=[4, 5]
  )
  test = dataset_info_pb2.SplitInfo(
      name="test", num_shards=3, shard_lengths=[1, 2, 3]
  )
  text_feature = feature_pb2.Feature(
      python_class_name="tensorflow_datasets.core.features.text_feature.Text",
      text=feature_pb2.TextFeature(),
  )
  proto = dataset_info_pb2.DatasetInfo(
      name="random_shaped_image_generator",
      version=str(generator_builder.version),
      features=feature_pb2.Feature(
          python_class_name=(
              "tensorflow_datasets.core.features.features_dict.FeaturesDict"
          ),
          features_dict=feature_pb2.FeaturesDict(
              features={"text": text_feature}
          ),
      ),
      splits=[train, test],
  )
  result = dataset_info.DatasetInfo.from_proto(
      builder=generator_builder, proto=proto
  )
  assert result.splits["test"].shard_lengths == test.shard_lengths
  assert result.splits["train"].shard_lengths == train.shard_lengths
  assert set(result.features.keys()) == {"text"}
  assert result.version == generator_builder.version


def test_supervised_keys_from_proto():
  proto = text_format.Parse(
      text="""
  tuple: {
      items: [
        {
          dict: {
            dict: {
              key: "f2"
              value: { feature_key: "f2" }
            },
            dict: {
              key: "f1"
              value: { feature_key: "f1" }
            },
          }
        },
        {
          feature_key: "target"
        }
      ]
    }
  """,
      message=dataset_info_pb2.SupervisedKeys(),
  )
  supervised_keys = dataset_info._supervised_keys_from_proto(proto=proto)
  assert str(supervised_keys) == "({'f1': 'f1', 'f2': 'f2'}, 'target')"


def test_supervised_keys_from_proto_different_ordering():
  proto1 = text_format.Parse(
      text="""
  tuple: {
      items: [
        {
          dict: {
            dict: {
              key: "f1"
              value: { feature_key: "f1" }
            },
            dict: {
              key: "f2"
              value: { feature_key: "f2" }
            },
            dict: {
              key: "f3"
              value: { feature_key: "f3" }
            },
          }
        },
        {
          feature_key: "target"
        }
      ]
    }
  """,
      message=dataset_info_pb2.SupervisedKeys(),
  )
  proto2 = text_format.Parse(
      text="""
  tuple: {
      items: [
        {
          dict: {
            dict: {
              key: "f3"
              value: { feature_key: "f3" }
            },
            dict: {
              key: "f2"
              value: { feature_key: "f2" }
            },
            dict: {
              key: "f1"
              value: { feature_key: "f1" }
            },
          }
        },
        {
          feature_key: "target"
        }
      ]
    }
  """,
      message=dataset_info_pb2.SupervisedKeys(),
  )
  supervised_keys1 = dataset_info._supervised_keys_from_proto(proto=proto1)
  supervised_keys2 = dataset_info._supervised_keys_from_proto(proto=proto2)
  assert str(supervised_keys1) == str(supervised_keys2)


@pytest.mark.parametrize(
    ("license_", "redistribution_info", "expected_license"),
    (
        ("a", None, "a"),
        (None, {"license": "a"}, "a"),
        ("a", {"license": "a"}, "a"),
        (None, None, None),
    ),
)
def test_create_redistribution_info_proto(
    license_, redistribution_info, expected_license
):
  result = dataset_info._create_redistribution_info_proto(
      license=license_, redistribution_info=redistribution_info
  )
  if expected_license is None:
    assert result is None
  else:
    assert result.license == expected_license


def test_create_redistribution_info_proto_inconsistent():
  with pytest.raises(
      ValueError, match="License specified twice and inconsistently"
  ):
    dataset_info._create_redistribution_info_proto(
        license="a", redistribution_info={"license": "b"}
    )


def test_create_redistribution_info_proto_unsupported_fields():
  with pytest.raises(
      ValueError, match="`redistribution_info` contains unsupported keys"
  ):
    dataset_info._create_redistribution_info_proto(
        license="a", redistribution_info={"unsupported": "b"}
    )


# pylint: disable=g-inconsistent-quotes
_INFO_STR = '''tfds.core.DatasetInfo(
    name='mnist',
    full_name='mnist/3.0.1',
    description="""
    The MNIST database of handwritten digits.
    """,
    homepage='https://storage.googleapis.com/cvdf-datasets/mnist/',
    data_dir='%s',
    file_format=tfrecord,
    download_size=1.95 KiB,
    dataset_size=11.06 MiB,
    features=FeaturesDict({
        'image': Image(shape=(28, 28, 1), dtype=uint8),
        'label': ClassLabel(shape=(), dtype=int64, num_classes=10),
    }),
    supervised_keys=('image', 'label'),
    disable_shuffling=False,
    splits={
        'test': <SplitInfo num_examples=20, num_shards=1>,
        'train': <SplitInfo num_examples=20, num_shards=1>,
    },
    citation="""@article{lecun2010mnist,
      title={MNIST handwritten digit database},
      author={LeCun, Yann and Cortes, Corinna and Burges, CJ},
      journal={ATT Labs [Online]. Available: http://yann. lecun. com/exdb/mnist},
      volume={2},
      year={2010}
    }
    """,
    redistribution_info=license: "test license",
)'''
# pylint: enable=g-inconsistent-quotes


if __name__ == "__main__":
  testing.test_main()



# coding=utf-8
# Copyright 2023 The TensorFlow Datasets Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Dataset feature for nested datasets."""
from __future__ import annotations

import functools
from typing import Any, Dict, Iterator, Union

from tensorflow_datasets.core.features import feature as feature_lib
from tensorflow_datasets.core.features import sequence_feature
from tensorflow_datasets.core.utils import py_utils
from tensorflow_datasets.core.utils import tree_utils
from tensorflow_datasets.core.utils import type_utils
from tensorflow_datasets.core.utils.lazy_imports_utils import tensorflow as tf


class Dataset(sequence_feature.Sequence):
  """A Dataset feature encodes a nested dataset.

  `Dataset` corresponds to a dataset of `tfds.features.FeatureConnector`. Using
  `tfds.features.Dataset` will return a nested `tf.data.Dataset` inside the
  top-level `tf.data.Dataset` returned by `tfds.load`. At generation time, an
  iterable over the dataset elements is given.

  This is an experimental feature. Currently, only one level of nesting is
  supported and TF1 graph is not supported either.

  Example:
  At construction time (inside `_info`):

  ```python
    features=tfds.features.FeatureDict({
     'agent_id': np.object_,
      'episode': tfds.features.Dataset({
        'observation': tfds.features.Image(),
        'reward': tfds.features.Image(),
      }),
    })
  ```


  Will return:

  ```
  {
    'agent_id': tf.Tensor(shape=(), dtype=tf.string),
    'episode': tf.data.Dataset(element_spec={
        'observation': tf.Tensor(shape=(None, None, 3), dtype=tf.uint8),
        'reward': tf.Tensor(shape=(), dtype=tf.int32),
    }),
    }
  ```

  The nested dataset can be used as:

  ```
  for e in tfds.load(...): # {'agent_id': tf.Tensor, 'episode': tf.data.Dataset}
    for step in e['episode']:  # Each episode is a nested  `tf.data.Dataset`
      step['observation']
  ```

  During generation,  it accept any `Iterable`/`Iterator`, like

  ```python
  yield _, {
    'agent_id': agent_name
    'episode': ({'observation': ..., 'reward': ...} for _ in range(10)),
  }
  ```

  Or a dictionary of `Iterable`, like

  ```python
  yield _, {
    'agent_id': agent_name
    'episode': {'observation': np.ones(10), 'reward': np.ones(10)} ,
  }
  ```
  """

  # TODO(tfds): Add support for TF1 graph mode.

  @py_utils.memoize()
  def get_tensor_info(self):
    """Shape of one element of the dataset."""
    # Add the dataset level
    tensor_info = self._feature.get_tensor_info()
    return tree_utils.map_structure(_add_dataset_lvl, tensor_info)

  def get_tensor_spec(self) -> tf.data.DatasetSpec:
    return tf.data.DatasetSpec(element_spec=self._feature.get_tensor_spec())

  @py_utils.memoize()
  def get_serialized_info(self):
    # Add the dataset level and the number of elements in the dataset
    tensor_info = super().get_serialized_info()
    return tree_utils.map_structure(_add_dataset_lvl, tensor_info)

  def encode_example(
      self,
      example_ds: Union[Iterator[type_utils.TreeDict[Any]], Dict[str, Any]],
  ):
    if isinstance(example_ds, dict):
      dict_list = sequence_feature.transpose_dict_list(example_ds)
    else:
      dict_list = example_ds
    # Encode each individual element
    ds_elements = [
        self.feature.encode_example(example) for example in dict_list
    ]

    # Empty datasets return empty arrays
    if not ds_elements:
      return tree_utils.map_structure(
          sequence_feature.build_empty_np, self.get_serialized_info()
      )

    # Then convert back list[nested dict] => nested dict[list]
    encoded = sequence_feature.stack_nested(ds_elements)
    return encoded

  def decode_example(self, serialized_example, decoders=None):
    # NOTE: By using from_tensor_slices we remove the possibility of nested
    # datasets.

    # Gets the decoding function of the inner feature to apply it to the
    # elements of the dataset.
    decode_fn = self.feature.decode_example
    if decoders:
      decode_fn = functools.partial(decode_fn, decoders=decoders)
    ds = tf.data.Dataset.from_tensor_slices(serialized_example).map(
        decode_fn, num_parallel_calls=tf.data.experimental.AUTOTUNE
    )
    return ds

  def _flatten(self, x):
    """See base class for details."""
    return [x]

  def _nest(self, list_x):
    """See base class for details."""
    assert len(list_x) == 1

    return list_x[0]


def _add_dataset_lvl(tensor_info):
  """Add the dataset nesting level to the tensor_info."""
  tensor_info = feature_lib.TensorInfo.copy_from(tensor_info)
  tensor_info.dataset_lvl += 1
  return tensor_info
